from collections import defaultdict
from dataclasses import dataclass, make_dataclass
from functools import cached_property
from os.path import basename
from typing import TYPE_CHECKING, Any, Optional, Union
from warnings import warn

from eth_typing import ChecksumAddress, HexAddress
from eth_utils import to_checksum_address

from vyper.semantics.analysis.base import FunctionVisibility, StateMutability
from vyper.utils import keccak256, method_id

from .abi import abi_decode, abi_encode, is_abi_encodable

if TYPE_CHECKING:
    from tests.evm_backends.base_env import BaseEnv, LogEntry


@dataclass
class ABILog:
    """Represents a parsed log entry from a contract."""

    address: ChecksumAddress
    args: tuple
    event: str
    topics: list[bytes]
    raw_data: bytes


class ABILogTopic:
    """Represents a log event topic in an ABI."""

    def __init__(self, event_abi: dict, contract_name: str):
        self._abi = event_abi
        self._contract_name = contract_name

    @cached_property
    def topic_id(self) -> bytes:
        """The keccak256 hash of the event signature."""
        if self._abi.get("anonymous") is True:
            return b""
        return keccak256((self.name + self.signature).encode())

    @property
    def name(self) -> str:
        return self._abi.get("name") or self._abi["type"]

    @property
    def indexed_inputs(self) -> list[dict]:
        return [item for item in self._abi["inputs"] if item["indexed"]]

    @property
    def unindexed_inputs(self) -> list[dict]:
        return [item for item in self._abi["inputs"] if not item["indexed"]]

    @cached_property
    def indexed_types(self) -> list[str]:
        return [_abi_from_json(i) for i in self.indexed_inputs]

    @cached_property
    def unindexed_types(self) -> list[str]:
        return [_abi_from_json(i) for i in self.unindexed_inputs]

    @property
    def signature(self) -> str:
        return f"({_format_abi_type(self.indexed_types + self.unindexed_types)})"

    def __repr__(self) -> str:
        return f"ABITopic {self._contract_name}.{self.signature} (0x{self.topic_id.hex()})"

    def parse(self, log: "LogEntry") -> ABILog:
        topics, raw_data = log.data
        return ABILog(
            address=to_checksum_address(log.address),
            args=self._parse_args(log),
            event=self.name,
            topics=topics,
            raw_data=raw_data,
        )

    @cached_property
    def data_type(self) -> type:
        """
        Create a dataclass for the log event data.
        """
        inputs = self.indexed_inputs + self.unindexed_inputs
        fields = [(item["name"], item["type"]) for item in inputs]
        return make_dataclass(self.name, fields)

    def _parse_args(self, log: "LogEntry") -> Any:
        """Convert the log data into a dataclass instance."""
        topics, data = log.data
        assert len(topics) == 1 + len(self.indexed_inputs), "Invalid log topic count"
        indexed = [
            t if self._is_hashed(typ) else abi_decode(f"{_format_abi_type([typ])}", t)
            for typ, t in zip(self.indexed_types, topics[1:])
        ]
        decoded = abi_decode(f"({_format_abi_type(self.unindexed_types)})", data)
        return self.data_type(*indexed, *decoded)

    @staticmethod
    def _is_hashed(typ):
        """Check if a type is hashed when included in a log topic."""
        return typ in ("bytes", "string", "tuple") or typ.endswith("[]")


class ABIFunction:
    """A single function in an ABI. It does not include overloads."""

    def __init__(self, abi: dict, contract_name: str):
        """
        :param abi: the ABI entry for this function
        :param contract_name: the name of the contract this function belongs to
        """
        self._abi = abi
        self._contract_name = contract_name
        self._function_visibility = FunctionVisibility.EXTERNAL
        self._mutability = StateMutability.from_abi(abi)
        self.contract: Optional["ABIContract"] = None

    @property
    def name(self) -> str:
        # note: the `constructor` definition does not have a name
        return self._abi.get("name") or self._abi["type"]

    @cached_property
    def argument_types(self) -> list:
        return [_abi_from_json(i) for i in self._abi["inputs"]]

    @property
    def argument_count(self) -> int:
        return len(self.argument_types)

    @property
    def _args_signature(self) -> str:
        return f"({_format_abi_type(self.argument_types)})"

    @cached_property
    def return_type(self) -> list:
        return [_abi_from_json(o) for o in self._abi["outputs"]]

    @property
    def full_signature(self) -> str:
        return f"{self.name}{self._args_signature}"

    @property
    def pretty_signature(self) -> str:
        return f"{self.name}{self._args_signature} -> {self.return_type}"

    @cached_property
    def method_id(self) -> bytes:
        if self._abi["type"] == "constructor":
            return b""  # constructors don't have method IDs
        return method_id(self.name + self._args_signature)

    def __repr__(self) -> str:
        return f"ABI {self._contract_name}.{self.pretty_signature}"

    def __str__(self) -> str:
        return repr(self)

    @property
    def is_mutable(self) -> bool:
        return self._mutability > StateMutability.VIEW

    def is_encodable(self, *args, **kwargs) -> bool:
        """Check whether this function accepts the given arguments after eventual encoding."""
        if len(kwargs) + len(args) != self.argument_count:
            return False
        parsed_args = self._merge_kwargs(*args, **kwargs)
        return all(
            is_abi_encodable(abi_type, arg)
            for abi_type, arg in zip(self.argument_types, parsed_args)
        )

    def prepare_calldata(self, *args, **kwargs) -> bytes:
        """Prepare the call data for the function call."""
        abi_args = self._merge_kwargs(*args, **kwargs)
        return self.method_id + abi_encode(self._args_signature, abi_args)

    def _merge_kwargs(self, *args, **kwargs) -> list:
        """Merge positional and keyword arguments into a single list."""
        if len(kwargs) + len(args) != self.argument_count:
            raise TypeError(
                "invocation failed due to improper number of arguments to"
                f" `{repr(self)}` (expected {self.argument_count} "
                f"arguments, got {len(args)} args and {len(kwargs)} kwargs)"
            )
        try:
            kwarg_inputs = self._abi["inputs"][len(args) :]
            return list(args) + [kwargs.pop(i["name"]) for i in kwarg_inputs]
        except KeyError as e:
            error = (
                f"Missing keyword argument {e} for `{self._args_signature}`. Passed {args} {kwargs}"
            )
            raise TypeError(error)

    def __call__(self, *args, value=0, gas=None, gas_price=0, sender=None, **kwargs):
        """Calls the function with the given arguments based on the ABI contract."""
        if not self.contract or not self.contract.env:
            raise Exception(f"Cannot call {self} without deploying contract.")

        if sender is None:
            sender = self.contract.env.deployer

        computation = self.contract.env.message_call(
            to=self.contract.address,
            sender=sender,
            data=self.prepare_calldata(*args, **kwargs),
            value=value,
            gas=gas,
            gas_price=gas_price,
            is_modifying=self.is_mutable,
        )

        match self.contract.marshal_to_python(computation, self.return_type):
            case ():
                return None
            case (single,):
                return single
            case multiple:
                return multiple


class ABIOverload:
    """
    Represents a set of functions that have the same name but different
    argument types. This is used to implement function overloading.
    """

    @staticmethod
    def create(
        functions: list[ABIFunction], contract: "ABIContract"
    ) -> Union["ABIOverload", ABIFunction]:
        """
        Create an ABIOverload if there are multiple functions, otherwise
        return the single function.
        :param functions: a list of functions with the same name
        :param contract: the ABIContract that these functions belong to
        """
        for f in functions:
            f.contract = contract
        if len(functions) == 1:
            return functions[0]
        return ABIOverload(functions)

    def __init__(self, functions: list[ABIFunction]):
        self.functions = functions

    @cached_property
    def name(self) -> str:
        return self.functions[0].name

    def __call__(
        self, *args, value=0, gas=None, sender=None, disambiguate_signature=None, **kwargs
    ):
        """
        Call the function that matches the given arguments.
        :raises Exception: if a single function is not found
        """
        function = self._pick_overload(
            *args, disambiguate_signature=disambiguate_signature, **kwargs
        )
        return function(*args, value=value, gas=gas, sender=sender, **kwargs)

    def _pick_overload(self, *args, disambiguate_signature=None, **kwargs) -> ABIFunction:
        """Pick the function that matches the given arguments."""
        if disambiguate_signature is None:
            matches = [f for f in self.functions if f.is_encodable(*args, **kwargs)]
        else:
            matches = [f for f in self.functions if disambiguate_signature == f.full_signature]
            assert len(matches) <= 1, "ABI signature must be unique"

        match matches:
            case [function]:
                return function
            case []:
                raise Exception(
                    f"Could not find matching {self.name} function for given arguments."
                )
            case multiple:
                raise Exception(
                    f"Ambiguous call to {self.name}. "
                    f"Arguments can be encoded to multiple overloads: "
                    f"{', '.join(self.name + f._args_signature for f in multiple)}. "
                    f"(Hint: try using `disambiguate_signature=` to disambiguate)."
                )


class ABIContract:
    """A contract that has been deployed to the blockchain and created via an ABI."""

    @property
    def address(self) -> HexAddress:
        assert self._address is not None
        return self._address

    def __init__(
        self,
        env: "BaseEnv",
        name: str,
        abi: dict,
        functions: list[ABIFunction],
        log_topics: list[ABILogTopic],
        bytecode: Optional[bytes],
        address: HexAddress,
        filename: Optional[str] = None,
    ):
        self.env = env
        self._address = address  # this can be overridden by subclasses
        self.filename = filename
        self.abi = abi
        self._name = name
        self._functions = functions
        self.log_topics = log_topics
        self.bytecode = bytecode
        self._deployed_bytecode = self.env.get_code(address)
        if not self._deployed_bytecode:
            warn(f"Requested {self} but there is no bytecode at that address!", stacklevel=2)

        overloads = defaultdict(list)
        for f in functions:
            overloads[f.name].append(f)

        for name, group in overloads.items():
            setattr(self, name, ABIOverload.create(group, self))

        self._address = address

    def marshal_to_python(self, result: bytes, abi_type: list[str]) -> list[Any]:
        """
        Convert the output of a contract call to a Python object.
        :param result: the computation result returned by `message_call`
        :param abi_type: the ABI type of the return value.
        """
        schema = f"({_format_abi_type(abi_type)})"
        return abi_decode(schema, result)

    def __repr__(self):
        file_str = f" (file {self.filename})" if self.filename else ""
        warn_str = "" if self._deployed_bytecode else " (WARNING: no bytecode at this address!)"
        return f"<{self._name} interface at {self.address}{warn_str}>{file_str}"

    def parse_log(self, log: "LogEntry") -> ABILog:
        """
        Parse a log entry into an ABILog object.
        :param log: the log entry to parse
        """
        topic_id_str = log.topics[0]
        topic_id = bytes.fromhex(topic_id_str.removeprefix("0x"))
        for topic in self.log_topics:
            if topic.topic_id == topic_id:
                return topic.parse(log)
        raise KeyError(f"Could not find event for log {topic_id_str}. Found {self.log_topics}")


class ABIContractFactory:
    """
    Represents an ABI contract that has not been coupled with an address yet.
    This is named `Factory` instead of `Deployer` because it doesn't actually
    do any contract deployment.
    """

    def __init__(
        self,
        name: str,
        abi: dict,
        functions: list[ABIFunction],
        log_topics: list[ABILogTopic],
        bytecode: Optional[bytes] = None,
    ):
        self._name = name
        self._abi = abi
        self._functions = functions
        self._log_topics = log_topics
        self._bytecode = bytecode

    @classmethod
    def from_abi_dict(cls, abi, name="<anonymous contract>", bytecode: Optional[bytes] = None):
        functions = [ABIFunction(item, name) for item in abi if item.get("type") == "function"]
        log_topics = [ABILogTopic(item, name) for item in abi if item.get("type") == "event"]
        return cls(basename(name), abi, functions, log_topics, bytecode)

    def at(self, env, address: HexAddress) -> ABIContract:
        """
        Create an ABI contract object for a deployed contract at `address`.
        """
        return ABIContract(
            env, self._name, self._abi, self._functions, self._log_topics, self._bytecode, address
        )


def _abi_from_json(abi: dict) -> str:
    """
    Parses an ABI type into its schema string.
    :param abi: The ABI type to parse.
    :return: The schema string for the given abi type.
    """
    if "components" in abi:
        components = ",".join([_abi_from_json(item) for item in abi["components"]])
        if abi["type"].startswith("tuple"):
            return f"({components}){abi['type'][5:]}"
        raise ValueError("Components found in non-tuple type " + abi["type"])

    return abi["type"]


def _format_abi_type(types: list) -> str:
    """
    Converts a list of ABI types into a comma-separated string.
    """
    return ",".join(
        item if isinstance(item, str) else f"({_format_abi_type(item)})" for item in types
    )



from . import primitives, subscriptable, user
from .base import TYPE_T, VOID_TYPE, KwargSettings, VyperType, is_type_t, map_void
from .bytestrings import BytesT, StringT, _BytestringT
from .function import MemberFunctionT
from .module import InterfaceT
from .primitives import AddressT, BoolT, BytesM_T, DecimalT, IntegerT, SelfT
from .subscriptable import DArrayT, HashMapT, SArrayT, TupleT
from .user import EventT, FlagT, StructT


def _get_primitive_types():
    res = [BoolT(), DecimalT()]

    res.extend(IntegerT.all())
    res.extend(BytesM_T.all())

    # order of the types matters!
    # parsing of literal hex: prefer address over bytes20
    res.append(AddressT())

    # note: since bytestrings are parametrizable, the *class* objects
    # are in the namespace instead of concrete type objects.
    res.extend([BytesT, StringT])

    ret = {t._id: t for t in res}
    ret.update(_get_sequence_types())

    return ret


def _get_sequence_types():
    # since these guys are parametrizable, the *class* objects
    # are in the namespace instead of concrete type objects.

    res = [HashMapT, DArrayT]

    ret = {t._id: t for t in res}

    # (static) arrays and tuples are special types which don't show up
    # in the type annotation itself.
    # since we don't have special handling of annotations in the parser,
    # break a dependency cycle by injecting these into the namespace with
    # mangled names (that no user can create).
    ret["$SArrayT"] = SArrayT
    ret["$TupleT"] = TupleT

    return ret


# note: it might be good to make this a frozen dict of some sort
PRIMITIVE_TYPES = _get_primitive_types()



import vyper


def test_basic_init_function(get_contract):
    code = """
val: public(uint256)

@deploy
def __init__(a: uint256):
    self.val = a
    """

    c = get_contract(code, *[123])

    assert c.val() == 123

    # Make sure the init code does not access calldata
    assembly = vyper.compile_code(code, output_formats=["asm"])["asm"].split(" ")
    ir_return_idx_start = assembly.index("{")
    ir_return_idx_end = assembly.index("}")

    assert "CALLDATALOAD" in assembly
    assert "CALLDATACOPY" not in assembly[:ir_return_idx_start] + assembly[ir_return_idx_end:]
    assert "CALLDATALOAD" not in assembly[:ir_return_idx_start] + assembly[ir_return_idx_end:]


def test_init_calls_internal(get_contract, assert_compile_failed, tx_failed):
    code = """
foo: public(uint8)

@internal
def bar(x: uint256) -> uint8:
    return convert(x, uint8) * 7

@deploy
def __init__(a: uint256):
    self.foo = self.bar(a)

@external
def baz() -> uint8:
    return self.bar(convert(self.foo, uint256))
    """
    n = 5
    c = get_contract(code, n)
    assert c.foo() == n * 7
    assert c.baz() == 245  # 5*7*7

    n = 6
    c = get_contract(code, n)
    assert c.foo() == n * 7
    with tx_failed():
        c.baz()

    n = 255
    assert_compile_failed(lambda: get_contract(code, n))

    n = 256
    assert_compile_failed(lambda: get_contract(code, n))


# GH issue 3206
def test_nested_internal_call_from_ctor(get_contract):
    code = """
x: uint256

@deploy
def __init__():
    self.a()

@internal
def a():
    self.x += 1
    self.b()

@internal
def b():
    self.x += 2

@external
def test() -> uint256:
    return self.x
    """
    c = get_contract(code)
    assert c.test() == 3



import re

import pytest

from vyper.compiler import compile_code
from vyper.exceptions import StorageLayoutException


def test_storage_layout_overrides():
    code = """
a: uint256
b: uint256"""

    storage_layout_overrides = {
        "a": {"type": "uint256", "slot": 1},
        "b": {"type": "uint256", "slot": 0},
    }

    expected_output = {"storage_layout": storage_layout_overrides, "code_layout": {}}

    out = compile_code(
        code, output_formats=["layout"], storage_layout_override=storage_layout_overrides
    )

    assert out["layout"] == expected_output


def test_storage_layout_for_more_complex():
    code = """
foo: HashMap[address, uint256]

@external
@nonreentrant
def public_foo1():
    pass

@external
@nonreentrant
def public_foo2():
    pass


@internal
@nonreentrant
def _bar():
    pass

# mix it up a little
baz: Bytes[65]
bar: uint256

@external
@nonreentrant
def public_bar():
    pass

@external
@nonreentrant
def public_foo3():
    pass
    """

    storage_layout_override = {
        "$.nonreentrant_key": {"type": "nonreentrant lock", "slot": 8},
        "foo": {"type": "HashMap[address, uint256]", "slot": 1},
        "baz": {"type": "Bytes[65]", "slot": 2},
        "bar": {"type": "uint256", "slot": 6},
    }

    expected_output = {"storage_layout": storage_layout_override, "code_layout": {}}

    out = compile_code(
        code, output_formats=["layout"], storage_layout_override=storage_layout_override
    )

    assert out["layout"] == expected_output


def test_simple_collision():
    code = """
name: public(String[64])
symbol: public(String[32])"""

    storage_layout_override = {
        "name": {"slot": 0, "type": "String[64]"},
        "symbol": {"slot": 1, "type": "String[32]"},
    }

    with pytest.raises(
        StorageLayoutException,
        match="Storage collision! Tried to assign 'symbol' to slot 1"
        " but it has already been reserved by 'name'",
    ):
        compile_code(
            code, output_formats=["layout"], storage_layout_override=storage_layout_override
        )


def test_overflow():
    code = """
x: uint256[2]
    """

    storage_layout_override = {"x": {"slot": 2**256 - 1, "type": "uint256[2]"}}

    with pytest.raises(
        StorageLayoutException, match=f"Invalid storage slot for var x, out of bounds: {2**256}"
    ):
        compile_code(
            code, output_formats=["layout"], storage_layout_override=storage_layout_override
        )


def test_override_nonreentrant_slot():
    code = """
@nonreentrant
@external
def foo():
    pass
    """

    storage_layout_override = {"$.nonreentrant_key": {"slot": 2**256, "type": "nonreentrant key"}}

    exception_regex = re.escape(
        f"Invalid storage slot for var $.nonreentrant_key, out of bounds: {2**256}"
    )
    with pytest.raises(StorageLayoutException, match=exception_regex):
        compile_code(
            code, output_formats=["layout"], storage_layout_override=storage_layout_override
        )


def test_incomplete_overrides():
    code = """
name: public(String[64])
symbol: public(String[32])"""

    storage_layout_override = {"name": {"slot": 0, "type": "String[64]"}}

    with pytest.raises(
        StorageLayoutException,
        match="Could not find storage_slot for symbol. "
        "Have you used the correct storage layout file?",
    ):
        compile_code(
            code, output_formats=["layout"], storage_layout_override=storage_layout_override
        )



from vyper.ast import parse_to_ast
from vyper.codegen.context import Context
from vyper.codegen.stmt import parse_body
from vyper.semantics.analysis.local import FunctionAnalyzer
from vyper.semantics.namespace import Namespace, override_global_namespace
from vyper.semantics.types.function import ContractFunctionT, FunctionVisibility, StateMutability
from vyper.semantics.types.module import ModuleT


def _strip_ast_source(ir_node):
    ir_node.ast_source = None
    for x in ir_node.args:
        _strip_ast_source(x)


def generate_inline_function(code, variables, variables_2, memory_allocator):
    ast_code = parse_to_ast(code, add_fn_node="dummy_fn")
    # Annotate the AST with a temporary old (i.e. typecheck) namespace
    namespace = Namespace()
    namespace.update(variables_2)
    with override_global_namespace(namespace):
        # Initialise a placeholder `FunctionDef` AST node and corresponding
        # `ContractFunctionT` type to rely on the annotation visitors in semantics
        # module.
        ast_code.body[0]._metadata["func_type"] = ContractFunctionT(
            "sqrt_builtin", [], [], None, FunctionVisibility.INTERNAL, StateMutability.NONPAYABLE
        )
        analyzer = FunctionAnalyzer(ast_code, ast_code.body[0], namespace)
        analyzer.analyze()

    new_context = Context(
        vars_=variables, module_ctx=ModuleT(ast_code), memory_allocator=memory_allocator
    )
    generated_ir = parse_body(ast_code.body[0].body, new_context)
    # strip source position info from the generated_ir since
    # it doesn't make any sense (e.g. the line numbers will start from 0
    # instead of where we are in the code)
    # NOTE if we ever use this for inlining user-code, it would make
    # sense to fix the offsets of the source positions in the generated
    # code instead of stripping them.
    _strip_ast_source(generated_ir)
    return new_context, generated_ir



import pytest

from vyper.ast import parse_to_ast
from vyper.exceptions import (
    ArrayIndexException,
    InvalidReference,
    TypeMismatch,
    UndeclaredDefinition,
)
from vyper.semantics.analysis import analyze_module


@pytest.mark.parametrize("value", ["address", "Bytes[10]", "decimal", "bool"])
def test_type_mismatch(namespace, value, dummy_input_bundle):
    code = f"""

a: uint256[3]

@internal
def foo(b: {value}):
    self.a[b] = 12
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(TypeMismatch):
        analyze_module(vyper_module, dummy_input_bundle)


@pytest.mark.parametrize("value", ["1.0", "0.0", "'foo'", "0x00", "b'\x01'", "False"])
def test_invalid_literal(namespace, value, dummy_input_bundle):
    code = f"""

a: uint256[3]

@internal
def foo():
    self.a[{value}] = 12
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(TypeMismatch):
        analyze_module(vyper_module, dummy_input_bundle)


@pytest.mark.parametrize("value", [-1, 3, -(2**127), 2**127 - 1, 2**256 - 1])
def test_out_of_bounds(namespace, value, dummy_input_bundle):
    code = f"""

a: uint256[3]

@internal
def foo():
    self.a[{value}] = 12
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(ArrayIndexException):
        analyze_module(vyper_module, dummy_input_bundle)


@pytest.mark.parametrize("value", ["b", "self.b"])
def test_undeclared_definition(namespace, value, dummy_input_bundle):
    code = f"""

a: uint256[3]

@internal
def foo():
    self.a[{value}] = 12
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(UndeclaredDefinition):
        analyze_module(vyper_module, dummy_input_bundle)


@pytest.mark.parametrize("value", ["a", "foo", "int128"])
def test_invalid_reference(namespace, value, dummy_input_bundle):
    code = f"""

a: uint256[3]

@internal
def foo():
    self.a[{value}] = 12
    """
    vyper_module = parse_to_ast(code)
    with pytest.raises(InvalidReference):
        analyze_module(vyper_module, dummy_input_bundle)



from pathlib import PurePath

import pytest

from vyper.cli.vyper_json import TRANSLATE_MAP, get_output_formats
from vyper.exceptions import JSONError


def test_no_outputs():
    with pytest.raises(KeyError):
        get_output_formats({})


def test_invalid_output():
    input_json = {
        "sources": {"foo.vy": ""},
        "settings": {"outputSelection": {"foo.vy": ["abi", "foobar"]}},
    }
    with pytest.raises(JSONError):
        get_output_formats(input_json)


def test_unknown_contract():
    input_json = {"sources": {}, "settings": {"outputSelection": {"bar.vy": ["abi"]}}}
    with pytest.raises(JSONError):
        get_output_formats(input_json)


@pytest.mark.parametrize("output", TRANSLATE_MAP.items())
def test_translate_map(output):
    input_json = {
        "sources": {"foo.vy": ""},
        "settings": {"outputSelection": {"foo.vy": [output[0]]}},
    }
    assert get_output_formats(input_json) == {PurePath("foo.vy"): [output[1]]}


def test_star():
    input_json = {
        "sources": {"foo.vy": "", "bar.vy": ""},
        "settings": {"outputSelection": {"*": ["*"]}},
    }
    expected = sorted(set(TRANSLATE_MAP.values()))
    result = get_output_formats(input_json)
    assert result == {PurePath("foo.vy"): expected, PurePath("bar.vy"): expected}


def test_ast():
    input_json = {
        "sources": {"foo.vy": ""},
        "settings": {"outputSelection": {"foo.vy": ["ast", "annotated_ast"]}},
    }
    expected = sorted([TRANSLATE_MAP[k] for k in ["ast", "annotated_ast"]])
    result = get_output_formats(input_json)
    assert result == {PurePath("foo.vy"): expected}


def test_evm():
    input_json = {
        "sources": {"foo.vy": ""},
        "settings": {"outputSelection": {"foo.vy": ["abi", "evm"]}},
    }
    expected = ["abi"] + sorted(v for k, v in TRANSLATE_MAP.items() if k.startswith("evm"))
    result = get_output_formats(input_json)
    assert result == {PurePath("foo.vy"): expected}


def test_solc_style():
    input_json = {
        "sources": {"foo.vy": ""},
        "settings": {"outputSelection": {"foo.vy": {"": ["abi"], "foo.vy": ["ir"]}}},
    }
    assert get_output_formats(input_json) == {PurePath("foo.vy"): ["abi", "ir_dict"]}


def test_metadata():
    input_json = {"sources": {"foo.vy": ""}, "settings": {"outputSelection": {"*": ["metadata"]}}}
    assert get_output_formats(input_json) == {PurePath("foo.vy"): ["metadata"]}



#!/usr/bin/env python3
import argparse
import sys

import vyper
from vyper.codegen.ir_node import IRnode
from vyper.ir import compile_ir, optimizer
from vyper.ir.s_expressions import parse_s_exp


def _parse_cli_args():
    return _parse_args(sys.argv[1:])


def _parse_args(argv):
    parser = argparse.ArgumentParser(description="Vyper IR IR compiler")
    parser.add_argument("input_file", help="Vyper sourcecode to compile")
    parser.add_argument(
        "--version", action="version", version=f"{vyper.__version__}+commit{vyper.__commit__}"
    )
    parser.add_argument(
        "-f",
        help="Format to print csv list of ir,opt_ir,asm,bytecode",
        default="bytecode",
        dest="format",
    )
    parser.add_argument(
        "--show-gas-estimates", help="Show gas estimates in ir output mode.", action="store_true"
    )

    args = parser.parse_args(argv)
    output_formats = set(dict.fromkeys(args.format.split(",")))
    compiler_data = compile_to_ir(args.input_file, output_formats, args.show_gas_estimates)

    for key in ("ir", "opt_ir", "asm", "bytecode"):
        if key in compiler_data:
            print(compiler_data[key])


def compile_to_ir(input_file, output_formats, show_gas_estimates=False):
    with open(input_file) as fh:
        s_expressions = parse_s_exp(fh.read())

    if show_gas_estimates:
        IRnode.repr_show_gas = True

    compiler_data = {}
    ir = IRnode.from_list(s_expressions[0])
    ir = optimizer.optimize(ir)
    if "ir" in output_formats:
        compiler_data["ir"] = ir

    asm = compile_ir.compile_to_assembly(ir)
    if "asm" in output_formats:
        compiler_data["asm"] = asm

    if "bytecode" in output_formats:
        bytecode, _ = compile_ir.assembly_to_evm(asm)
        compiler_data["bytecode"] = "0x" + bytecode.hex()

    return compiler_data


if __name__ == "__main__":
    _parse_cli_args()



import pytest

from vyper.semantics.types.utils import type_from_annotation

BASE_TYPES = ["int128", "uint256", "bool", "address", "bytes32"]
BYTESTRING_TYPES = ["String", "Bytes"]


@pytest.mark.parametrize("type_str", BASE_TYPES)
def test_base_types(build_node, type_str):
    node = build_node(type_str)
    type_definition = type_from_annotation(node)

    assert type_definition.size_in_bytes == 32


@pytest.mark.parametrize("type_str", BYTESTRING_TYPES)
@pytest.mark.parametrize("length,size", [(1, 64), (32, 64), (33, 96), (86, 128)])
def test_array_value_types(build_node, type_str, length, size):
    node = build_node(f"{type_str}[{length}]")
    type_definition = type_from_annotation(node)

    assert type_definition.size_in_bytes == size


@pytest.mark.parametrize("type_str", BASE_TYPES)
@pytest.mark.parametrize("length", range(1, 4))
def test_dynamic_array_lengths(build_node, type_str, length):
    node = build_node(f"DynArray[{type_str}, {length}]")
    type_definition = type_from_annotation(node)

    assert type_definition.size_in_bytes == 32 + length * 32


@pytest.mark.parametrize("type_str", BASE_TYPES)
@pytest.mark.parametrize("length", range(1, 4))
def test_base_types_as_arrays(build_node, type_str, length):
    node = build_node(f"{type_str}[{length}]")
    type_definition = type_from_annotation(node)

    assert type_definition.size_in_bytes == length * 32


@pytest.mark.parametrize("type_str", BASE_TYPES)
@pytest.mark.parametrize("first", range(1, 4))
@pytest.mark.parametrize("second", range(1, 4))
def test_base_types_as_multidimensional_arrays(build_node, type_str, first, second):
    node = build_node(f"{type_str}[{first}][{second}]")

    type_definition = type_from_annotation(node)

    assert type_definition.size_in_bytes == first * second * 32



import hypothesis.strategies as st
import pytest
from hypothesis import given, settings

from vyper.compiler import compile_code
from vyper.compiler.settings import OptimizationLevel, Settings
from vyper.evm.opcodes import version_check
from vyper.exceptions import ArgumentException, TypeMismatch

_fun_bytes32_bounds = [(0, 32), (3, 29), (27, 5), (0, 5), (5, 3), (30, 2)]


def _generate_bytes(length):
    return bytes(list(range(length)))


def test_basic_slice(get_contract):
    code = """
@external
def slice_tower_test(inp1: Bytes[50]) -> Bytes[50]:
    inp: Bytes[50] = inp1
    for i: uint256 in range(1, 11):
        inp = slice(inp, 1, 30 - i * 2)
    return inp
    """
    c = get_contract(code)
    x = c.slice_tower_test(b"abcdefghijklmnopqrstuvwxyz1234")
    assert x == b"klmnopqrst", x


# note: optimization boundaries at 32, 64 and 320 depending on mode
_draw_1024 = st.integers(min_value=0, max_value=1024)
_draw_1024_1 = st.integers(min_value=1, max_value=1024)
_bytes_1024 = st.binary(min_size=0, max_size=1024)


def _fail_contract(code, opt_level, exceptions):
    settings = Settings(optimize=opt_level)
    with pytest.raises(exceptions):
        compile_code(code, settings=settings)


@pytest.mark.parametrize("use_literal_start", (True, False))
@pytest.mark.parametrize("use_literal_length", (True, False))
@pytest.mark.parametrize("opt_level", list(OptimizationLevel))
@given(start=_draw_1024, length=_draw_1024, length_bound=_draw_1024_1, bytesdata=_bytes_1024)
@settings(max_examples=100)
@pytest.mark.fuzzing
def test_slice_immutable(
    get_contract,
    tx_failed,
    opt_level,
    bytesdata,
    start,
    use_literal_start,
    length,
    use_literal_length,
    length_bound,
):
    _start = start if use_literal_start else "start"
    _length = length if use_literal_length else "length"

    code = f"""
IMMUTABLE_BYTES: immutable(Bytes[{length_bound}])
IMMUTABLE_SLICE: immutable(Bytes[{length_bound}])

@deploy
def __init__(inp: Bytes[{length_bound}], start: uint256, length: uint256):
    IMMUTABLE_BYTES = inp
    IMMUTABLE_SLICE = slice(IMMUTABLE_BYTES, {_start}, {_length})

@external
def do_splice() -> Bytes[{length_bound}]:
    return IMMUTABLE_SLICE
    """

    def _get_contract():
        return get_contract(code, bytesdata, start, length, override_opt_level=opt_level)

    if (
        (start + length > length_bound and use_literal_start and use_literal_length)
        or (use_literal_length and length > length_bound)
        or (use_literal_start and start > length_bound)
        or (use_literal_length and length == 0)
    ):
        _fail_contract(code, opt_level, ArgumentException)

    elif start + length > len(bytesdata) or (len(bytesdata) > length_bound):
        # deploy fail
        with tx_failed():
            _get_contract()
    else:
        c = _get_contract()
        assert c.do_splice() == bytesdata[start : start + length]


@pytest.mark.parametrize(
    "location", ["storage", "transient", "calldata", "memory", "literal", "code"]
)
@pytest.mark.parametrize("use_literal_start", (True, False))
@pytest.mark.parametrize("use_literal_length", (True, False))
@pytest.mark.parametrize("opt_level", list(OptimizationLevel))
@given(start=_draw_1024, length=_draw_1024, length_bound=_draw_1024_1, bytesdata=_bytes_1024)
@settings(max_examples=100)
@pytest.mark.fuzzing
def test_slice_bytes_fuzz(
    get_contract,
    tx_failed,
    opt_level,
    location,
    bytesdata,
    start,
    use_literal_start,
    length,
    use_literal_length,
    length_bound,
):
    if location == "transient" and not version_check(begin="cancun"):
        pytest.skip(
            "Skipping test as storage_location is 'transient' and EVM version is pre-Cancun"
        )
    preamble = ""
    if location == "memory":
        spliced_code = f"foo: Bytes[{length_bound}] = inp"
        foo = "foo"
    elif location == "storage":
        preamble = f"""
foo: Bytes[{length_bound}]
         """
        spliced_code = "self.foo = inp"
        foo = "self.foo"
    elif location == "transient":
        preamble = f"""
foo: transient(Bytes[{length_bound}])
        """
        spliced_code = "self.foo = inp"
        foo = "self.foo"
    elif location == "code":
        preamble = f"""
IMMUTABLE_BYTES: immutable(Bytes[{length_bound}])
@deploy
def __init__(foo: Bytes[{length_bound}]):
    IMMUTABLE_BYTES = foo
    """
        spliced_code = ""
        foo = "IMMUTABLE_BYTES"
    elif location == "literal":
        spliced_code = ""
        foo = f"{bytesdata}"
    elif location == "calldata":
        spliced_code = ""
        foo = "inp"
    else:
        raise Exception("unreachable")

    _start = start if use_literal_start else "start"
    _length = length if use_literal_length else "length"

    code = f"""
{preamble}

@external
def do_slice(inp: Bytes[{length_bound}], start: uint256, length: uint256) -> Bytes[{length_bound}]:
    {spliced_code}
    return slice({foo}, {_start}, {_length})
    """

    def _get_contract():
        if "__init__" in code:
            # eth-tester used to ignore constructor arguments if no constructor was defined
            # now we raise an exception, so only call the constructor if it exists
            # TODO: Refactor so we don't rely on searching the source code.
            return get_contract(code, bytesdata, override_opt_level=opt_level)
        return get_contract(code, override_opt_level=opt_level)

    # length bound is the container size; input_bound is the bound on the input
    # (which can be different, if the input is a literal)
    input_bound = length_bound
    slice_output_too_large = False

    if location == "literal":
        input_bound = len(bytesdata)

        # ex.:
        # @external
        # def do_slice(inp: Bytes[1], start: uint256, length: uint256) -> Bytes[1]:
        #    return slice(b'\x00\x00', 0, length)
        output_length = length if use_literal_length else input_bound
        slice_output_too_large = output_length > length_bound

    end = start + length

    compile_time_oob = (
        (use_literal_length and (length > input_bound or length == 0))
        or (use_literal_start and start > input_bound)
        or (use_literal_start and use_literal_length and start + length > input_bound)
    )

    if compile_time_oob or slice_output_too_large:
        _fail_contract(code, opt_level, (ArgumentException, TypeMismatch))

    elif location == "code" and len(bytesdata) > length_bound:
        # deploy fail
        with tx_failed():
            _get_contract()
    elif end > len(bytesdata) or len(bytesdata) > length_bound:
        c = _get_contract()
        with tx_failed():
            c.do_slice(bytesdata, start, length)
    else:
        c = _get_contract()
        assert c.do_slice(bytesdata, start, length) == bytesdata[start:end], code


@pytest.mark.parametrize("location", ["storage", "transient"])
def test_slice_private(get_contract, location):
    if location == "transient" and not version_check(begin="cancun"):
        pytest.skip(
            "Skipping test as storage_location is 'transient' and EVM version is pre-Cancun"
        )

    # test there are no buffer overruns in the slice function
    if location == "storage":
        decl = "bytez: public(String[12])"
    elif location == "transient":
        decl = "bytez: public(transient(String[12]))"
    else:
        raise Exception("unreachable")

    code = f"""
{decl}

@internal
def _slice(start: uint256, length: uint256):
    self.bytez = slice(self.bytez, start, length)

@external
def foo(x: uint256, y: uint256) -> (uint256, String[12]):
    self.bytez = "hello, world"
    dont_clobber_me: uint256 = max_value(uint256)
    self._slice(x, y)
    return dont_clobber_me, self.bytez
    """
    c = get_contract(code)
    assert c.foo(0, 12) == (2**256 - 1, "hello, world")
    assert c.foo(12, 0) == (2**256 - 1, "")
    assert c.foo(7, 5) == (2**256 - 1, "world")
    assert c.foo(0, 5) == (2**256 - 1, "hello")
    assert c.foo(0, 1) == (2**256 - 1, "h")
    assert c.foo(11, 1) == (2**256 - 1, "d")


def test_slice_storage_bytes32(get_contract):
    code = """
bytez: bytes32
@external
def dice() -> Bytes[1]:
    self.bytez = convert(65, bytes32)
    c: Bytes[1] = slice(self.bytez, 31, 1)
    return c
    """

    c = get_contract(code)
    assert c.dice() == b"A"


def test_slice_immutable_length_arg(get_contract):
    code = """
LENGTH: immutable(uint256)

@deploy
def __init__():
    LENGTH = 5

@external
def do_slice(inp: Bytes[50]) -> Bytes[50]:
    return slice(inp, 0, LENGTH)
    """
    c = get_contract(code)
    x = c.do_slice(b"abcdefghijklmnopqrstuvwxyz1234")
    assert x == b"abcde", x


def test_slice_at_end(get_contract):
    code = """
@external
def ret10_slice() -> Bytes[10]:
    b: Bytes[32] = concat(convert(65, bytes32), b'')
    c: Bytes[10] = slice(b, 31, 1)
    return c
    """

    c = get_contract(code)
    assert c.ret10_slice() == b"A"


def test_slice_expr(get_contract):
    # test slice of a complex expression
    code = """
@external
def ret10_slice() -> Bytes[10]:
    return slice(convert(65, bytes32), 31, 1)
    """

    c = get_contract(code)
    assert c.ret10_slice() == b"A"


def test_slice_equality(get_contract):
    # test for equality with dirty bytes
    code = """
@external
def assert_eq() -> bool:
    dirty_bytes: String[4] = "abcd"
    dirty_bytes = slice(dirty_bytes, 0, 3)
    clean_bytes: String[4] = "abc"
    return dirty_bytes == clean_bytes
    """

    c = get_contract(code)
    assert c.assert_eq()


def test_slice_inequality(get_contract):
    # test for equality with dirty bytes
    code = """
@external
def assert_ne() -> bool:
    dirty_bytes: String[4] = "abcd"
    dirty_bytes = slice(dirty_bytes, 0, 3)
    clean_bytes: String[4] = "abcd"
    return dirty_bytes != clean_bytes
    """

    c = get_contract(code)
    assert c.assert_ne()


def test_slice_convert(get_contract):
    # test slice of converting between bytes32 and Bytes
    code = """
@external
def f() -> bytes32:
    a: Bytes[100] = convert("ab", Bytes[100])
    return convert(slice(a, 0, 1), bytes32)
    """
    c = get_contract(code)
    assert c.f() == b"a" + b"\x00" * 31


code_bytes32 = [
    """
foo: bytes32

@deploy
def __init__():
    self.foo = 0x000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f

@external
def bar() -> Bytes[{length}]:
    return slice(self.foo, {start}, {length})
    """,
    """
foo: bytes32

@deploy
def __init__():
    self.foo = 0x000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f

@external
def bar() -> Bytes[32]:
    a: uint256 = {start}
    b: uint256 = {length}
    return slice(self.foo, a, b)
    """,
    f"""
foo: Bytes[32]

@external
def bar() -> Bytes[32]:
    self.foo = {_generate_bytes(32)}
    a: uint256 = {{start}}
    b: uint256 = {{length}}
    return slice(convert(self.foo, bytes32), a, b)
    """,
    """
@external
def bar() -> Bytes[{length}]:
    foo: bytes32 = 0x000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f
    return slice(foo, {start}, {length})
    """,
    """
@external
def bar() -> Bytes[32]:
    b: uint256 = {length}
    foo: bytes32 = 0x000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f
    a: uint256 = {start}
    return slice(foo, a, b)
    """,
]


@pytest.mark.parametrize("code", code_bytes32)
@pytest.mark.parametrize("start,length", _fun_bytes32_bounds)
def test_slice_bytes32(get_contract, code, start, length):
    c = get_contract(code.format(start=start, length=length))
    assert c.bar() == _generate_bytes(32)[start : start + length]


code_bytes32_calldata = [
    """
@external
def bar(foo: bytes32) -> Bytes[{length}]:
    return slice(foo, {start}, {length})
    """,
    """
@external
def bar(foo: bytes32) -> Bytes[32]:
    b: uint256 = {length}
    a: uint256 = {start}
    return slice(foo, a, b)
    """,
]


@pytest.mark.parametrize("code", code_bytes32_calldata)
@pytest.mark.parametrize("start,length", _fun_bytes32_bounds)
def test_slice_bytes32_calldata(get_contract, code, start, length):
    c = get_contract(code.format(start=start, length=length))
    assert c.bar(_generate_bytes(32)) == _generate_bytes(32)[start : start + length]


code_bytes32_calldata_extended = [
    (
        """
@external
def bar(a: uint256, foo: bytes32, b: uint256) -> Bytes[32]:
    return slice(foo, 3, 5)
    """,
        "0304050607",
    ),
    (
        """
@external
def bar(a: uint256, foo: bytes32, b: uint256) -> Bytes[32]:
    return slice(foo, a, b)
    """,
        "0304050607",
    ),
    (
        """
@external
def bar(a: uint256, foo: bytes32, b: uint256) -> Bytes[32]:
    return slice(foo, 31, b-4)
    """,
        "31",
    ),
    (
        """
@external
def bar(a: uint256, foo: bytes32, b: uint256) -> Bytes[32]:
    return slice(foo, 0, a+b)
    """,
        "0001020304050607",
    ),
]


@pytest.mark.parametrize("code,result", code_bytes32_calldata_extended)
def test_slice_bytes32_calldata_extended(get_contract, code, result):
    c = get_contract(code)
    assert (
        c.bar(3, "0x0001020304050607080910111213141516171819202122232425262728293031", 5).hex()
        == result
    )


# test cases crafted based on advisory GHSA-9x7f-gwxq-6f2c
oob_fail_list = [
    """
d: public(Bytes[256])

@external
def do_slice():
    x : uint256 = max_value(uint256)
    self.d = b"\x01\x02\x03\x04\x05\x06"
    assert len(slice(self.d, 1, x)) == max_value(uint256)
    """,
    """
@external
def do_slice():
    x: uint256 = max_value(uint256)
    # y == 0x3232323232323232323232323232323232323232323232323232323232323232
    y: uint256 = 22704331223003175573249212746801550559464702875615796870481879217237868556850
    z: uint96 = 1
    if True:
        placeholder : uint256[16] = [y, y, y, y, y, y, y, y, y, y, y, y, y, y, y, y]
    s: String[32] = slice(uint2str(z), 1, x)
    assert slice(s, 1, 2) == "22"
    """,
    """
x: public(Bytes[64])
secret: uint256

@deploy
def __init__():
    self.x = empty(Bytes[64])
    self.secret = 42

@external
def do_slice() -> Bytes[64]:
    start: uint256 = max_value(uint256) - 63
    return slice(self.x, start, 64)
    """,
    # tests bounds check in adhoc location calldata
    """
interface IFace:
    def choose_value(_x: uint256, _y: uint256, _z: uint256, idx: uint256) -> Bytes[32]: nonpayable

@external
def choose_value(_x: uint256, _y: uint256, _z: uint256, idx: uint256) -> Bytes[32]:
    assert idx % 32 == 4
    return slice(msg.data, idx, 32)

@external
def do_slice():
    idx: uint256 = max_value(uint256) - 27
    ret: uint256 = _abi_decode(extcall IFace(self).choose_value(1, 2, 3, idx), uint256)
    assert ret == 0
    """,
]


@pytest.mark.parametrize("bad_code", oob_fail_list)
def test_slice_buffer_oob_reverts(bad_code, get_contract, tx_failed):
    c = get_contract(bad_code)
    with tx_failed():
        c.do_slice()



import pytest

from vyper import compile_code
from vyper.exceptions import TypeMismatch

fail_list = [
    (  # bad AST nodes given as arguments
        """
@external
def foo() -> uint256:
    return uint256_addmod(1.1, 1.2, 3.0)
    """,
        TypeMismatch,
    ),
    (  # bad AST nodes given as arguments
        """
@external
def foo() -> uint256:
    return uint256_mulmod(1.1, 1.2, 3.0)
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("code,exc", fail_list)
def test_add_mod_fail(assert_compile_failed, get_contract, code, exc):
    assert_compile_failed(lambda: get_contract(code), exc)


valid_list = [
    """
FOO: constant(uint256) = 3
BAR: constant(uint256) = 5
BAZ: constant(uint256) = 19
BAX: constant(uint256) = uint256_addmod(FOO, BAR, BAZ)
    """,
    """
FOO: constant(uint256) = 3
BAR: constant(uint256) = 5
BAZ: constant(uint256) = 19
BAX: constant(uint256) = uint256_mulmod(FOO, BAR, BAZ)
    """,
]


@pytest.mark.parametrize("code", valid_list)
def test_addmulmod_pass(code):
    assert compile_code(code) is not None



import itertools
import operator
import random

import pytest

from vyper.semantics.types import IntegerT
from vyper.utils import evm_div, unsigned_to_signed

# TODO something less janky
integer_types = sorted(list(IntegerT.all()))


@pytest.mark.parametrize("typ", integer_types)
@pytest.mark.parametrize("op", ["add", "sub", "mul", "div"])
@pytest.mark.fuzzing
def test_unsafe_op_int(get_contract, typ, op):
    contract_1 = f"""
@external
def foo(x: {typ}, y: {typ}) -> {typ}:
    return unsafe_{op}(x, y)
    """

    contract_2 = """
@external
def foo(x: {typ}) -> {typ}:
    return unsafe_{op}(x, {literal})
    """

    fns = {"add": operator.add, "sub": operator.sub, "mul": operator.mul, "div": evm_div}
    fn = fns[op]

    c1 = get_contract(contract_1)

    lo, hi = typ.ast_bounds
    # (roughly 8k cases total generated)
    # TODO refactor to use fixtures
    NUM_CASES = 15
    xs = [random.randrange(lo, hi) for _ in range(NUM_CASES)]
    ys = [random.randrange(lo, hi) for _ in range(NUM_CASES)]

    mod_bound = 2**typ.bits

    # poor man's fuzzing - hypothesis doesn't make it easy
    # with the parametrized strategy
    if typ.is_signed:
        xs += [lo, lo + 1, -1, 0, 1, hi - 1, hi]
        ys += [lo, lo + 1, -1, 0, 1, hi - 1, hi]
        for x, y in itertools.product(xs, ys):
            expected = unsigned_to_signed(fn(x, y) % mod_bound, typ.bits)

            assert c1.foo(x, y) == expected

            c2 = get_contract(contract_2.format(typ=typ, op=op, literal=y))
            assert c2.foo(x) == expected

    else:
        # 0x80 has some weird properties, like
        # it's a fixed point of multiplication by 0xFF
        fixed_pt = 2 ** (typ.bits - 1)
        xs += [0, 1, hi - 1, hi, fixed_pt]
        ys += [0, 1, hi - 1, hi, fixed_pt]
        for x, y in itertools.product(xs, ys):
            expected = fn(x, y) % mod_bound
            assert c1.foo(x, y) == expected

            c2 = get_contract(contract_2.format(typ=typ, op=op, literal=y))
            assert c2.foo(x) == expected



import pytest

from vyper import compiler
from vyper.exceptions import StructureException

FAILING_CONTRACTS = [
    """
@external
@pure
@nonreentrant
def nonreentrant_foo() -> uint256:
    return 1
    """,
    """
@external
@nonreentrant
@nonreentrant
def nonreentrant_foo() -> uint256:
    return 1
    """,
    """
@external
@nonreentrant("foo")
def nonreentrant_foo() -> uint256:
    return 1
    """,
]


@pytest.mark.parametrize("failing_contract_code", FAILING_CONTRACTS)
def test_invalid_function_decorators(failing_contract_code):
    with pytest.raises(StructureException):
        compiler.compile_code(failing_contract_code)



import operator
from typing import List, Optional, Tuple, Union

from vyper.codegen.ir_node import IRnode
from vyper.evm.opcodes import version_check
from vyper.exceptions import CompilerPanic, StaticAssertionException
from vyper.utils import (
    ceil32,
    evm_div,
    evm_mod,
    evm_pow,
    int_bounds,
    int_log2,
    is_power_of_two,
    signed_to_unsigned,
    unsigned_to_signed,
)

SIGNED = False
UNSIGNED = True


# unsigned: convert python num to evm unsigned word
#   e.g. unsigned=True : -1 -> 0xFF...FF
#        unsigned=False: 0xFF...FF -> -1
def _evm_int(node: IRnode, unsigned: bool = True) -> Optional[int]:
    if isinstance(node.value, int):
        ret = node.value
    else:
        return None

    if unsigned and ret < 0:
        return signed_to_unsigned(ret, 256, strict=True)
    elif not unsigned and ret > 2**255 - 1:
        return unsigned_to_signed(ret, 256, strict=True)

    return ret


def _is_int(node: IRnode) -> bool:
    return isinstance(node.value, int)


def _deep_contains(node_or_list, node):
    if isinstance(node_or_list, list):
        return any(_deep_contains(t, node) for t in node_or_list)
    return node is node_or_list


arith = {
    "add": (operator.add, "+", UNSIGNED),
    "sub": (operator.sub, "-", UNSIGNED),
    "mul": (operator.mul, "*", UNSIGNED),
    "div": (evm_div, "/", UNSIGNED),
    "sdiv": (evm_div, "/", SIGNED),
    "mod": (evm_mod, "%", UNSIGNED),
    "smod": (evm_mod, "%", SIGNED),
    "exp": (evm_pow, "**", UNSIGNED),
    "eq": (operator.eq, "==", UNSIGNED),
    "ne": (operator.ne, "!=", UNSIGNED),
    "lt": (operator.lt, "<", UNSIGNED),
    "le": (operator.le, "<=", UNSIGNED),
    "gt": (operator.gt, ">", UNSIGNED),
    "ge": (operator.ge, ">=", UNSIGNED),
    "slt": (operator.lt, "<", SIGNED),
    "sle": (operator.le, "<=", SIGNED),
    "sgt": (operator.gt, ">", SIGNED),
    "sge": (operator.ge, ">=", SIGNED),
    "or": (operator.or_, "|", UNSIGNED),
    "and": (operator.and_, "&", UNSIGNED),
    "xor": (operator.xor, "^", UNSIGNED),
}

# quick typedefs, maybe move these to IRnode
IRVal = Union[str, int]
IRArgs = List[IRnode]


COMMUTATIVE_OPS = {"add", "mul", "eq", "ne", "and", "or", "xor"}
COMPARISON_OPS = {"gt", "sgt", "ge", "sge", "lt", "slt", "le", "sle"}
STRICT_COMPARISON_OPS = {t for t in COMPARISON_OPS if t.endswith("t")}
UNSTRICT_COMPARISON_OPS = {t for t in COMPARISON_OPS if t.endswith("e")}

assert not (STRICT_COMPARISON_OPS & UNSTRICT_COMPARISON_OPS)
assert STRICT_COMPARISON_OPS | UNSTRICT_COMPARISON_OPS == COMPARISON_OPS


def _flip_comparison_op(opname):
    assert opname in COMPARISON_OPS
    if "g" in opname:
        return opname.replace("g", "l")
    if "l" in opname:
        return opname.replace("l", "g")
    raise CompilerPanic(f"bad comparison op {opname}")  # pragma: nocover


# some annotations are really long. shorten them (except maybe in "verbose" mode?)
def _shorten_annotation(annotation):
    if len(annotation) > 16:
        return annotation[:16] + "..."
    return annotation


def _wrap256(x, unsigned=UNSIGNED):
    x %= 2**256
    # wrap in a signed way.
    if not unsigned:
        x = unsigned_to_signed(x, 256, strict=True)
    return x


def _comparison_helper(binop, args, prefer_strict=False):
    assert binop in COMPARISON_OPS

    if _is_int(args[0]):
        binop = _flip_comparison_op(binop)
        args = [args[1], args[0]]

    unsigned = not binop.startswith("s")
    is_strict = binop.endswith("t")
    is_gt = "g" in binop

    # local version of _evm_int which defaults to the current binop's signedness
    def _int(x):
        return _evm_int(x, unsigned=unsigned)

    lo, hi = int_bounds(bits=256, signed=not unsigned)

    # for comparison operators, we have three special boundary cases:
    # almost always, never and almost never.
    # almost_always is always true for the non-strict ("ge" and co)
    # comparators. for strict comparators ("gt" and co), almost_always
    # is true except for one case. never is never true for the strict
    # comparators. never is almost always false for the non-strict
    # comparators, except for one case. and almost_never is almost
    # never true (except one case) for the strict comparators.
    if is_gt:
        almost_always, never = lo, hi
        almost_never = hi - 1
    else:
        almost_always, never = hi, lo
        almost_never = lo + 1

    if is_strict and _int(args[1]) == never:
        # e.g. gt x MAX_UINT256, slt x MIN_INT256
        return (0, [])

    if not is_strict and _int(args[1]) == almost_always:
        # e.g. ge x MIN_UINT256, sle x MAX_INT256
        return (1, [])

    if is_strict and _int(args[1]) == almost_never:
        # (lt x 1), (gt x (MAX_UINT256 - 1)), (slt x (MIN_INT256 + 1))
        return ("eq", [args[0], never])

    # rewrites. in positions where iszero is preferred, (gt x 5) => (ge x 6)
    if is_strict != prefer_strict and _is_int(args[1]):
        rhs = _int(args[1])

        if prefer_strict and rhs == never:
            # e.g. ge x MAX_UINT256, sle x MIN_INT256
            return ("eq", args)

        if not prefer_strict and rhs == almost_always:
            # e.g. gt x 0, slt x MAX_INT256
            return ("ne", args)

        if is_gt == is_strict:
            # x > 1 => x >= 2
            # x <= 1 => x < 2
            new_rhs = rhs + 1
        else:
            # x >= 1 => x > 0
            # x < 1 => x <= 0
            new_rhs = rhs - 1

        # if args[1] is OOB, it should have been handled above
        # in the always/never cases
        assert _wrap256(new_rhs, unsigned) == new_rhs, "bad optimizer step"

        # change the strictness of the op
        if prefer_strict:
            # e.g. "sge" => "sgt"
            new_op = binop.replace("e", "t")
        else:
            # e.g. "sgt" => "sge"
            new_op = binop.replace("t", "e")

        return (new_op, [args[0], new_rhs])

    # special cases that are not covered by others:

    if binop == "gt" and _int(args[1]) == 0:
        # improve codesize (not gas), and maybe trigger
        # downstream optimizations
        return ("iszero", [["iszero", args[0]]])


# def _optimize_arith(
#    binop: str, args: IRArgs, ann: Optional[str], parent_op: Any = None
# ) -> Tuple[IRVal, IRArgs, Optional[str]]:
def _optimize_binop(binop, args, ann, parent_op):
    fn, symb, unsigned = arith[binop]

    # local version of _evm_int which defaults to the current binop's signedness
    def _int(x, unsigned=unsigned):
        return _evm_int(x, unsigned=unsigned)

    def _wrap(x):
        return _wrap256(x, unsigned=unsigned)

    new_ann = None
    if ann is not None:
        l_ann = _shorten_annotation(args[0].annotation or str(args[0]))
        r_ann = _shorten_annotation(args[1].annotation or str(args[1]))
        new_ann = l_ann + symb + r_ann
        new_ann = f"{ann} ({new_ann})"

    def finalize(new_val, new_args):
        # if the original had side effects which might not be in the
        # optimized output, roll back the optimization
        rollback = (args[0].is_complex_ir and not _deep_contains(new_args, args[0])) or (
            args[1].is_complex_ir and not _deep_contains(new_args, args[1])
        )

        if rollback:
            return None

        return new_val, new_args, new_ann

    if _is_int(args[0]) and _is_int(args[1]):
        # compile-time arithmetic
        left, right = _int(args[0]), _int(args[1])
        new_val = fn(left, right)
        # wrap the result, since `fn` generally does not wrap.
        # (note: do not rely on wrapping/non-wrapping behavior for `fn`!
        # some ops, like evm_pow, ALWAYS wrap).
        new_val = _wrap(new_val)
        return finalize(new_val, [])

    # we can return truthy values instead of actual math
    is_truthy = parent_op in {"if", "assert", "iszero"}

    def _conservative_eq(x, y):
        # whether x evaluates to the same value as y at runtime.
        # TODO we can do better than this check, but we need to be
        # conservative in case x has side effects.
        return x.args == y.args == [] and x.value == y.value and not x.is_complex_ir

    ##
    # ARITHMETIC AND BITWISE OPS
    ##

    # for commutative ops, move the literal to the second
    # position to make the later logic cleaner
    if binop in COMMUTATIVE_OPS and _is_int(args[0]):
        args = [args[1], args[0]]

    if binop in {"add", "sub", "xor", "or"} and _int(args[1]) == 0:
        # x + 0 == x - 0 == x | 0 == x ^ 0 == x
        return finalize("seq", [args[0]])

    if binop in {"sub", "xor", "ne"} and _conservative_eq(args[0], args[1]):
        # (x - x) == (x ^ x) == (x != x) == 0
        return finalize(0, [])

    if binop in STRICT_COMPARISON_OPS and _conservative_eq(args[0], args[1]):
        # (x < x) == (x > x) == 0
        return finalize(0, [])

    if binop in {"eq"} | UNSTRICT_COMPARISON_OPS and _conservative_eq(args[0], args[1]):
        # (x == x) == (x >= x) == (x <= x) == 1
        return finalize(1, [])

    # TODO associativity rules

    # x * 0 == x / 0 == x % 0 == x & 0 == 0
    if binop in {"mul", "div", "sdiv", "mod", "smod", "and"} and _int(args[1]) == 0:
        return finalize(0, [])

    # x % 1 == 0
    if binop in {"mod", "smod"} and _int(args[1]) == 1:
        return finalize(0, [])

    # x * 1 == x / 1 == x
    if binop in {"mul", "div", "sdiv"} and _int(args[1]) == 1:
        return finalize("seq", [args[0]])

    # x * -1 == 0 - x
    if binop in {"mul", "sdiv"} and _int(args[1], SIGNED) == -1:
        return finalize("sub", [0, args[0]])

    if binop in {"and", "or", "xor"} and _int(args[1], SIGNED) == -1:
        assert unsigned == UNSIGNED
        if binop == "and":
            # -1 & x == x
            return finalize("seq", [args[0]])

        if binop == "xor":
            # -1 ^ x == ~x
            return finalize("not", [args[0]])

        if binop == "or":
            # -1 | x == -1
            return finalize(args[1].value, [])

        raise CompilerPanic("unreachable")  # pragma: nocover

    # -1 - x == ~x (definition of two's complement)
    if binop == "sub" and _int(args[0], SIGNED) == -1:
        return finalize("not", [args[1]])

    if binop == "exp":
        # n ** 0 == 1 (forall n)
        # 1 ** n == 1
        if _int(args[1]) == 0 or _int(args[0]) == 1:
            return finalize(1, [])
        # 0 ** n == (1 if n == 0 else 0)
        if _int(args[0]) == 0:
            return finalize("iszero", [args[1]])
        # n ** 1 == n
        if _int(args[1]) == 1:
            return finalize("seq", [args[0]])

    # TODO: check me! reduce codesize for negative numbers
    # if binop in {"add", "sub"} and _int(args[1], SIGNED) < 0:
    #     flipped = "add" if binop == "sub" else "sub"
    #     return finalize(flipped, [args[0], -args[1]])

    # TODO maybe OK:
    # elif binop == "div" and _int(args[1], UNSIGNED) == MAX_UINT256:
    #    # (div x (2**256 - 1)) == (eq x (2**256 - 1))
    #    new_val = "eq"
    #    args = args

    if binop in {"mod", "div", "mul"} and _is_int(args[1]) and is_power_of_two(_int(args[1])):
        assert unsigned == UNSIGNED, "something's not right."
        # shave two gas off mod/div/mul for powers of two
        # x % 2**n == x & (2**n - 1)
        if binop == "mod":
            return finalize("and", [args[0], _int(args[1]) - 1])

        if binop == "div":
            # x / 2**n == x >> n
            # recall shr/shl have unintuitive arg order
            return finalize("shr", [int_log2(_int(args[1])), args[0]])

        # note: no rule for sdiv since it rounds differently from sar
        if binop == "mul":
            # x * 2**n == x << n
            return finalize("shl", [int_log2(_int(args[1])), args[0]])

        raise CompilerPanic("unreachable")  # pragma: no cover

    ##
    # COMPARISONS
    ##

    if binop == "eq" and _int(args[1]) == 0:
        return finalize("iszero", [args[0]])

    # can't improve gas but can improve codesize
    if binop == "ne" and _int(args[1]) == 0:
        return finalize("iszero", [["iszero", args[0]]])

    if binop == "eq" and _int(args[1], SIGNED) == -1:
        # equal gas, but better codesize
        # x == MAX_UINT256 => ~x == 0
        return finalize("iszero", [["not", args[0]]])

    # note: in places where truthy is accepted, sequences of
    # ISZERO ISZERO will be optimized out, so we try to rewrite
    # some operations to include iszero
    # (note ordering; truthy optimizations should come first
    # to avoid getting clobbered by other branches)
    if is_truthy:
        if binop == "eq":
            assert unsigned == UNSIGNED
            # (eq x y) has the same truthyness as (iszero (xor x y))
            # it also has the same truthyness as (iszero (sub x y)),
            # but xor is slightly easier to optimize because of being
            # commutative.
            # note that (xor (-1) x) has its own rule
            return finalize("iszero", [["xor", args[0], args[1]]])

        if binop == "ne" and parent_op == "iszero":
            # for iszero, trigger other optimizations
            # (for `if` and `assert`, `ne` will generate two ISZEROs
            # which will get optimized out during assembly)
            return finalize("iszero", [["eq", *args]])

        # TODO can we do this?
        # if val == "div":
        #     return finalize("gt", ["iszero", args])

        if binop == "or" and _is_int(args[1]) and _int(args[1]) != 0:
            # (x | y != 0) for any (y != 0)
            return finalize(1, [])

    if binop in COMPARISON_OPS:
        prefer_strict = not is_truthy
        res = _comparison_helper(binop, args, prefer_strict=prefer_strict)
        if res is None:
            return res
        new_op, new_args = res
        return finalize(new_op, new_args)

    # no optimization happened
    return None


def _check_symbols(symbols, ir_node):
    # sanity check that no `unique_symbol`s got optimized out.
    to_check = ir_node.unique_symbols
    if symbols != to_check:
        raise CompilerPanic(f"missing symbols: {symbols - to_check}")


def optimize(node: IRnode) -> IRnode:
    _, ret = _optimize(node, parent=None)
    return ret


def _optimize(node: IRnode, parent: Optional[IRnode]) -> Tuple[bool, IRnode]:
    starting_symbols = node.unique_symbols

    res = [_optimize(arg, node) for arg in node.args]
    argz: list
    if len(res) == 0:
        args_changed, argz = False, []
    else:
        changed_flags, argz = zip(*res)  # type: ignore
        args_changed = any(changed_flags)
        argz = list(argz)

    value = node.value
    typ = node.typ
    location = node.location
    ast_source = node.ast_source
    error_msg = node.error_msg
    annotation = node.annotation
    add_gas_estimate = node.add_gas_estimate
    is_self_call = node.is_self_call
    passthrough_metadata = node.passthrough_metadata

    changed = False

    # in general, we cannot enforce the symbols check. for instance,
    # the dead branch eliminator will almost always trip the symbols check.
    # but for certain operations, particularly binops, we want to do the check.
    should_check_symbols = False

    def finalize(val, args):
        if not changed and not args_changed:
            # skip IRnode.from_list, which may be (compile-time) expensive
            return False, node

        ir_builder = [val, *args]
        ret = IRnode.from_list(
            ir_builder,
            typ=typ,
            location=location,
            ast_source=ast_source,
            error_msg=error_msg,
            annotation=annotation,
            add_gas_estimate=add_gas_estimate,
            is_self_call=is_self_call,
            passthrough_metadata=passthrough_metadata,
        )

        if should_check_symbols:
            _check_symbols(starting_symbols, ret)

        _, ret = _optimize(ret, parent)
        return True, ret

    if value == "seq":
        changed |= _merge_memzero(argz)
        changed |= _merge_calldataload(argz)
        changed |= _merge_dload(argz)
        changed |= _rewrite_mstore_dload(argz)
        changed |= _merge_mload(argz)
        changed |= _remove_empty_seqs(argz)

        # (seq x) => (x) for cleanliness and
        # to avoid blocking other optimizations
        if len(argz) == 1:
            return True, _optimize(argz[0], parent)[1]

        return finalize(value, argz)

    if value in arith:
        parent_op = parent.value if parent is not None else None

        res = _optimize_binop(value, argz, annotation, parent_op)
        if res is not None:
            changed = True
            should_check_symbols = True
            value, argz, annotation = res  # type: ignore
            return finalize(value, argz)

    ###
    # BITWISE OPS
    ###

    # note, don't optimize these too much as these kinds of expressions
    # may be hand optimized for codesize. we can optimize bitwise ops
    # more, once we have a pipeline which optimizes for codesize.
    if value in ("shl", "shr", "sar") and argz[0].value == 0:
        # x >> 0 == x << 0 == x
        changed = True
        annotation = argz[1].annotation
        return finalize(argz[1].value, argz[1].args)

    if node.value == "ceil32" and _is_int(argz[0]):
        changed = True
        annotation = f"ceil32({argz[0].value})"
        return finalize(ceil32(argz[0].value), [])

    if value == "iszero" and _is_int(argz[0]):
        changed = True
        val = int(argz[0].value == 0)  # int(bool) == 1 if bool else 0
        return finalize(val, [])

    if node.value == "if":
        # optimize out the branch
        if _is_int(argz[0]):
            changed = True
            # if false
            if _evm_int(argz[0]) == 0:
                # return the else branch (or [] if there is no else)
                return finalize("seq", argz[2:])
            # if true
            else:
                # return the first branch
                return finalize("seq", [argz[1]])

        elif len(argz) == 3 and argz[0].value not in ("iszero", "ne"):
            # if(x) compiles to jumpi(_, iszero(x))
            # there is an asm optimization for the sequence ISZERO ISZERO..JUMPI
            # so we swap the branches here to activate that optimization.
            cond = argz[0]
            true_branch = argz[1]
            false_branch = argz[2]
            contra_cond = IRnode.from_list(["iszero", cond])

            argz = [contra_cond, false_branch, true_branch]
            changed = True
            return finalize("if", argz)

    if value in ("assert", "assert_unreachable") and _is_int(argz[0]):
        if _evm_int(argz[0]) == 0:
            raise StaticAssertionException(
                f"assertion found to fail at compile time. (hint: did you mean `raise`?) {node}",
                ast_source,
            )
        else:
            changed = True
            return finalize("seq", [])

    return finalize(value, argz)


def _merge_memzero(argz):
    # look for sequential mzero / calldatacopy operations that are zero'ing memory
    # and merge them into a single calldatacopy
    mstore_nodes: List = []
    initial_offset = 0
    total_length = 0
    changed = False
    idx = None
    for i, ir_node in enumerate(argz):
        is_last_iteration = i == len(argz) - 1

        if (
            ir_node.value == "mstore"
            and isinstance(ir_node.args[0].value, int)
            and ir_node.args[1].value == 0
        ):
            # mstore of a zero value
            offset = ir_node.args[0].value
            if not mstore_nodes:
                idx = i
                initial_offset = offset
            if initial_offset + total_length == offset:
                mstore_nodes.append(ir_node)
                total_length += 32
                # do not block the optimization if it continues thru
                # the end of the (seq) block
                if not is_last_iteration:
                    continue

        if (
            ir_node.value == "calldatacopy"
            and isinstance(ir_node.args[0].value, int)
            and ir_node.args[1].value == "calldatasize"
            and isinstance(ir_node.args[2].value, int)
        ):
            # calldatacopy from the end of calldata - efficient zero'ing via `empty()`
            offset, length = ir_node.args[0].value, ir_node.args[2].value
            if not mstore_nodes:
                idx = i
                initial_offset = offset
            if initial_offset + total_length == offset:
                mstore_nodes.append(ir_node)
                total_length += length
                # do not block the optimization if it continues thru
                # the end of the (seq) block
                if not is_last_iteration:
                    continue

        # if we get this far, the current node is not a zero'ing operation
        # it's time to apply the optimization if possible
        if len(mstore_nodes) > 1:
            changed = True
            new_ir = IRnode.from_list(
                ["calldatacopy", initial_offset, "calldatasize", total_length],
                ast_source=mstore_nodes[0].ast_source,
            )
            # replace first zero'ing operation with optimized node and remove the rest
            argz[idx] = new_ir
            # note: del xs[k:l] deletes l - k items
            del argz[idx + 1 : idx + len(mstore_nodes)]

        initial_offset = 0
        total_length = 0
        mstore_nodes.clear()

    return changed


# remove things like [seq, seq, seq] because they interfere with
# other optimizer steps
def _remove_empty_seqs(argz):
    changed = False
    i = 0
    while i < len(argz) - 1:
        if argz[i].value in ("seq", "pass") and len(argz[i].args) == 0:
            changed = True
            del argz[i]
        else:
            i += 1
    return changed


def _merge_calldataload(argz):
    return _merge_load(argz, "calldataload", "calldatacopy")


def _merge_dload(argz):
    return _merge_load(argz, "dload", "dloadbytes")


def _rewrite_mstore_dload(argz):
    changed = False
    for i, arg in enumerate(argz):
        if arg.value == "mstore" and arg.args[1].value == "dload":
            dst = arg.args[0]
            src = arg.args[1].args[0]
            len_ = 32
            argz[i] = IRnode.from_list(["dloadbytes", dst, src, len_], ast_source=arg.ast_source)
            changed = True
    return changed


def _merge_mload(argz):
    if not version_check(begin="cancun"):
        return False
    return _merge_load(argz, "mload", "mcopy", allow_overlap=False)


def _merge_load(argz, _LOAD, _COPY, allow_overlap=True):
    # look for sequential operations copying from X to Y
    # and merge them into a single copy operation
    changed = False
    mstore_nodes: List = []
    initial_dst_offset = 0
    initial_src_offset = 0
    total_length = 0
    idx = None
    for i, ir_node in enumerate(argz):
        is_last_iteration = i == len(argz) - 1
        if (
            ir_node.value == "mstore"
            and isinstance(ir_node.args[0].value, int)
            and ir_node.args[1].value == _LOAD
            and isinstance(ir_node.args[1].args[0].value, int)
        ):
            # mstore of a zero value
            dst_offset = ir_node.args[0].value
            src_offset = ir_node.args[1].args[0].value
            if not mstore_nodes:
                initial_dst_offset = dst_offset
                initial_src_offset = src_offset
                idx = i

            # dst and src overlap, discontinue the optimization
            has_overlap = initial_src_offset < initial_dst_offset < src_offset + 32

            if (
                initial_dst_offset + total_length == dst_offset
                and initial_src_offset + total_length == src_offset
                and (allow_overlap or not has_overlap)
            ):
                mstore_nodes.append(ir_node)
                total_length += 32

                # do not block the optimization if it continues thru
                # the end of the (seq) block
                if not is_last_iteration:
                    continue

        # if we get this far, the current node is a different operation
        # it's time to apply the optimization if possible
        if len(mstore_nodes) > 1:
            changed = True
            new_ir = IRnode.from_list(
                [_COPY, initial_dst_offset, initial_src_offset, total_length],
                ast_source=mstore_nodes[0].ast_source,
            )
            # replace first copy operation with optimized node and remove the rest
            argz[idx] = new_ir
            # note: del xs[k:l] deletes l - k items
            del argz[idx + 1 : idx + len(mstore_nodes)]

        initial_dst_offset = 0
        initial_src_offset = 0
        total_length = 0
        mstore_nodes.clear()

    return changed



from vyper.codegen.abi_encoder import abi_encode
from vyper.codegen.core import get_type_for_exact_size, ir_tuple_from_args, unwrap_location
from vyper.codegen.ir_node import IRnode
from vyper.codegen.keccak256_helper import keccak256_helper
from vyper.exceptions import TypeMismatch
from vyper.semantics.types.bytestrings import _BytestringT


# docs.soliditylang.org/en/v0.8.6/abi-spec.html#indexed-event-encoding
def _encode_log_topics(expr, event_id, arg_nodes, context):
    topics = [event_id]

    for arg in arg_nodes:
        if arg.typ._is_prim_word:
            value = unwrap_location(arg)

        elif isinstance(arg.typ, _BytestringT):
            value = keccak256_helper(arg, context=context)
        else:
            # TODO block at higher level
            raise TypeMismatch("Event indexes may only be value types", expr)

        topics.append(value)

    return topics


def _gas_bound(num_topics, data_maxlen):
    LOG_BASE_GAS = 375
    GAS_PER_TOPIC = 375
    GAS_PER_LOG_BYTE = 8
    return LOG_BASE_GAS + GAS_PER_TOPIC * num_topics + GAS_PER_LOG_BYTE * data_maxlen


# docs.soliditylang.org/en/v0.8.6/abi-spec.html#events
def ir_node_for_log(expr, event, topic_nodes, data_nodes, context):
    """Taking IR nodes as arguments, create the IR node for a Log statement.

    Arguments:
      expr: The original Log expression
      buf: A pre-allocated buffer for the output
      _maxlen: The length of the buffer, for sanity checking
      event: The Event type
      topic_nodes: list of IRnodes which calculate the event topics
      data_nodes: list of IRnodes which calculate the event data
      context: current memory/frame context
    """
    topics = _encode_log_topics(expr, event.event_id, topic_nodes, context)

    data = ir_tuple_from_args(data_nodes)

    bufsz = data.typ.abi_type.size_bound()
    buf = context.new_internal_variable(get_type_for_exact_size(bufsz))

    # encode_data is an IRnode which, cleverly, both encodes the data
    # and returns the length of the encoded data as a stack item.
    encode_data = abi_encode(buf, data, context, returns_len=True, bufsz=bufsz)

    assert len(topics) <= 4, "too many topics"  # sanity check
    log_opcode = "log" + str(len(topics))

    return IRnode.from_list(
        [log_opcode, buf, encode_data] + topics,
        add_gas_estimate=_gas_bound(len(topics), bufsz),
        annotation=f"LOG event {event.signature}",
    )



import json
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Callable, Optional

from eth_keys.datatypes import PrivateKey
from eth_utils import to_checksum_address

from tests.evm_backends.abi import abi_decode
from tests.evm_backends.abi_contract import ABIContract, ABIContractFactory, ABIFunction
from vyper.ast.grammar import parse_vyper_source
from vyper.compiler import CompilerData, InputBundle, Settings, compile_code
from vyper.utils import ERC5202_PREFIX, method_id


# a very simple log representation for the raw log entries
@dataclass
class LogEntry:
    address: str
    topics: list[str]
    data: tuple[list[bytes], bytes]  # (topic list, non-topic)


# object returned by `last_result` property
@dataclass
class ExecutionResult:
    is_success: bool
    logs: list[LogEntry]
    gas_refunded: int
    gas_used: int


class EvmError(RuntimeError):
    """Exception raised when a call fails."""


class ExecutionReverted(EvmError):
    """Exception raised when a call reverts."""


class BaseEnv:
    """
    Base class for EVM backends.
    It provides a common interface for deploying contracts and interacting with them.
    """

    DEFAULT_CHAIN_ID = 1

    def __init__(self, gas_limit: int, account_keys: list[PrivateKey]) -> None:
        self.gas_limit = gas_limit
        self._keys = account_keys
        self.deployer: str = self._keys[0].public_key.to_checksum_address()

    def deploy(self, abi: list[dict], bytecode: bytes, value=0, *args, **kwargs):
        """Deploy a contract with the given ABI and bytecode."""
        factory = ABIContractFactory.from_abi_dict(abi, bytecode=bytecode)

        initcode = bytecode
        if args or kwargs:
            ctor_abi = next(i for i in abi if i["type"] == "constructor")
            ctor = ABIFunction(ctor_abi, contract_name=factory._name)
            initcode += ctor.prepare_calldata(*args, **kwargs)

        deployed_at = self._deploy(initcode, value)
        address = to_checksum_address(deployed_at)
        return factory.at(self, address)

    def deploy_source(
        self,
        source_code: str,
        output_formats: dict[str, Callable[[CompilerData], str]],
        *args,
        compiler_settings: Settings = None,
        input_bundle: InputBundle = None,
        value: int = 0,
        **kwargs,
    ) -> ABIContract:
        """Compile and deploy a contract from source code."""
        abi, bytecode = _compile(source_code, output_formats, input_bundle, compiler_settings)
        return self.deploy(abi, bytecode, value, *args, **kwargs)

    def deploy_blueprint(
        self,
        source_code,
        output_formats,
        *args,
        input_bundle: InputBundle = None,
        initcode_prefix: bytes = ERC5202_PREFIX,
    ):
        """Deploy a contract with a blueprint pattern."""
        abi, bytecode = _compile(source_code, output_formats, input_bundle)
        bytecode = initcode_prefix + bytecode
        bytecode_len = len(bytecode)
        bytecode_len_hex = hex(bytecode_len)[2:].rjust(4, "0")
        # prepend a quick deploy preamble
        deploy_preamble = bytes.fromhex("61" + bytecode_len_hex + "3d81600a3d39f3")
        deploy_bytecode = deploy_preamble + bytecode

        deployer_abi: list[dict] = []  # just a constructor
        deployer = self.deploy(deployer_abi, deploy_bytecode, *args)

        def factory(address):
            return ABIContractFactory.from_abi_dict(abi).at(self, address)

        return deployer, factory

    def get_logs(self, contract: ABIContract, event_name: str = None, raw=False):
        logs = [log for log in self.last_result.logs if contract.address == log.address]
        if raw:
            return [log.data for log in logs]

        parsed_logs = [contract.parse_log(log) for log in logs]
        if event_name:
            return [log for log in parsed_logs if log.event == event_name]

        return parsed_logs

    @property
    def accounts(self) -> list[str]:
        return [key.public_key.to_checksum_address() for key in self._keys]

    @contextmanager
    def anchor(self):
        raise NotImplementedError  # must be implemented by subclasses

    def get_balance(self, address: str) -> int:
        raise NotImplementedError  # must be implemented by subclasses

    def set_balance(self, address: str, value: int):
        raise NotImplementedError  # must be implemented by subclasses

    @property
    def block_number(self) -> int:
        raise NotImplementedError  # must be implemented by subclasses

    @block_number.setter
    def block_number(self, value: int):
        raise NotImplementedError

    @property
    def timestamp(self) -> int | None:
        raise NotImplementedError  # must be implemented by subclasses

    @timestamp.setter
    def timestamp(self, value: int):
        raise NotImplementedError  # must be implemented by subclasses

    @property
    def last_result(self) -> ExecutionResult:
        raise NotImplementedError  # must be implemented by subclasses

    @property
    def blob_hashes(self) -> list[bytes]:
        raise NotImplementedError  # must be implemented by subclasses

    @blob_hashes.setter
    def blob_hashes(self, value: list[bytes]):
        raise NotImplementedError  # must be implemented by subclasses

    def message_call(
        self,
        to: str,
        sender: str | None = None,
        data: bytes | str = b"",
        value: int = 0,
        gas: int | None = None,
        gas_price: int = 0,
        is_modifying: bool = True,
    ) -> bytes:
        raise NotImplementedError  # must be implemented by subclasses

    def clear_transient_storage(self) -> None:
        raise NotImplementedError  # must be implemented by subclasses

    def get_code(self, address: str) -> bytes:
        raise NotImplementedError  # must be implemented by subclasses

    def get_excess_blob_gas(self) -> Optional[int]:
        raise NotImplementedError  # must be implemented by subclasses

    def set_excess_blob_gas(self, param):
        raise NotImplementedError  # must be implemented by subclasses

    def _deploy(self, code: bytes, value: int, gas: int | None = None) -> str:
        raise NotImplementedError  # must be implemented by subclasses

    @staticmethod
    def _parse_revert(output_bytes: bytes, error: Exception, gas_used: int):
        """
        Tries to parse the EIP-838 revert reason from the output bytes.
        """
        if output_bytes[:4] == method_id("Error(string)"):
            (msg,) = abi_decode("(string)", output_bytes[4:])
            raise ExecutionReverted(f"{msg}", gas_used) from error

        raise ExecutionReverted(f"0x{output_bytes.hex()}", gas_used) from error

    @property
    def invalid_opcode_error(self) -> str:
        """Expected error message when invalid opcode is executed."""
        raise NotImplementedError  # must be implemented by subclasses

    @property
    def out_of_gas_error(self) -> str:
        """Expected error message when user runs out of gas"""
        raise NotImplementedError  # must be implemented by subclasses


def _compile(
    source_code: str,
    output_formats: dict[str, Callable[[CompilerData], str]],
    input_bundle: InputBundle = None,
    settings: Settings = None,
) -> tuple[list[dict], bytes]:
    out = compile_code(
        source_code,
        # test that all output formats can get generated
        output_formats=output_formats,
        settings=settings,
        input_bundle=input_bundle,
        show_gas_estimates=True,  # Enable gas estimates for testing
    )
    parse_vyper_source(source_code)  # Test grammar.
    json.dumps(out["metadata"])  # test metadata is json serializable
    return out["abi"], bytes.fromhex(out["bytecode"].removeprefix("0x"))



import pytest

from vyper.exceptions import InvalidType, UnknownType

fail_list = [
    """
x: bat
    """,
    """
x: HashMap[int, int128]
    """,
    """
struct A:
    b: B
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_unknown_type_exception(bad_code, get_contract, assert_compile_failed):
    assert_compile_failed(lambda: get_contract(bad_code), UnknownType)


invalid_list = [
    # Must be a literal string.
    """
@external
def mint(_to: address, _value: uint256):
    assert msg.sender == self,msg.sender
    """,
    # Raise reason must be string
    """
@external
def mint(_to: address, _value: uint256):
    raise 1
    """,
    """
x: int128[3.5]
    """,
    # Key of mapping must be a base type
    """
b: HashMap[(int128, decimal), int128]
    """,
    """
x: String <= 33
    """,
    """
x: Bytes <= wei
    """,
    """
x: 5
    """,
]


@pytest.mark.parametrize("bad_code", invalid_list)
def test_invalid_type_exception(bad_code, get_contract, assert_compile_failed):
    assert_compile_failed(lambda: get_contract(bad_code), InvalidType)



def test_import_interface_types(make_input_bundle, get_contract):
    ifaces = """
interface IFoo:
    def foo() -> uint256: nonpayable
    """

    foo_impl = """
import ifaces

implements: ifaces.IFoo

@external
def foo() -> uint256:
    return block.number
    """

    contract = """
import ifaces

@external
def test_foo(s: ifaces.IFoo) -> bool:
    assert extcall s.foo() == block.number
    return True
    """

    input_bundle = make_input_bundle({"ifaces.vy": ifaces})

    foo = get_contract(foo_impl, input_bundle=input_bundle)
    c = get_contract(contract, input_bundle=input_bundle)

    assert c.test_foo(foo.address) is True


def test_import_interface_types_stability(make_input_bundle, get_contract):
    lib1 = """
from ethereum.ercs import IERC20
    """
    lib2 = """
from ethereum.ercs import IERC20
    """

    main = """
import lib1
import lib2

from ethereum.ercs import IERC20

@external
def foo() -> bool:
    # check that this typechecks both directions
    a: lib1.IERC20 = IERC20(msg.sender)
    b: lib2.IERC20 = IERC20(msg.sender)

    # return the equality so we can sanity check it
    return a == b
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(main, input_bundle=input_bundle)

    assert c.foo() is True



import vyper.codegen.events as events
import vyper.utils as util
from vyper import ast as vy_ast
from vyper.codegen.abi_encoder import abi_encode
from vyper.codegen.context import Constancy, Context
from vyper.codegen.core import (
    LOAD,
    STORE,
    IRnode,
    add_ofst,
    clamp_le,
    get_dyn_array_count,
    get_element_ptr,
    get_type_for_exact_size,
    make_setter,
    wrap_value_for_external_return,
)
from vyper.codegen.expr import Expr
from vyper.codegen.return_ import make_return_stmt
from vyper.evm.address_space import MEMORY, STORAGE
from vyper.exceptions import CodegenPanic, StructureException, TypeCheckFailure, tag_exceptions
from vyper.semantics.types import DArrayT
from vyper.semantics.types.shortcuts import UINT256_T


class Stmt:
    def __init__(self, node: vy_ast.VyperNode, context: Context) -> None:
        self.stmt = node
        self.context = context

        fn_name = f"parse_{type(node).__name__}"
        with tag_exceptions(node, fallback_exception_type=CodegenPanic, note=fn_name):
            fn = getattr(self, fn_name)
            with context.internal_memory_scope():
                self.ir_node = fn()

            assert isinstance(self.ir_node, IRnode), self.ir_node

        self.ir_node.annotation = self.stmt.get("node_source_code")
        self.ir_node.ast_source = self.stmt

    def parse_Expr(self):
        return Expr(self.stmt.value, self.context, is_stmt=True).ir_node

    def parse_Pass(self):
        return IRnode.from_list("pass")

    def parse_Name(self):
        if self.stmt.id == "vdb":
            return IRnode("debugger")
        else:
            raise StructureException(f"Unsupported statement type: {type(self.stmt)}", self.stmt)

    def parse_AnnAssign(self):
        ltyp = self.stmt.target._metadata["type"]
        varname = self.stmt.target.id
        alloced = self.context.new_variable(varname, ltyp)

        assert self.stmt.value is not None
        rhs = Expr(self.stmt.value, self.context).ir_node

        lhs = IRnode.from_list(alloced, typ=ltyp, location=MEMORY)

        return make_setter(lhs, rhs)

    def parse_Assign(self):
        # Assignment (e.g. x[4] = y)
        src = Expr(self.stmt.value, self.context).ir_node
        dst = self._get_target(self.stmt.target)

        ret = ["seq"]
        overlap = len(dst.referenced_variables & src.referenced_variables) > 0
        if overlap and not dst.typ._is_prim_word:
            # there is overlap between the lhs and rhs, and the type is
            # complex - i.e., it spans multiple words. for safety, we
            # copy to a temporary buffer before copying to the destination.
            tmp = self.context.new_internal_variable(src.typ)
            tmp = IRnode.from_list(tmp, typ=src.typ, location=MEMORY)
            ret.append(make_setter(tmp, src))
            src = tmp

        ret.append(make_setter(dst, src))
        return IRnode.from_list(ret)

    def parse_If(self):
        with self.context.block_scope():
            test_expr = Expr.parse_value_expr(self.stmt.test, self.context)
            body = ["if", test_expr, parse_body(self.stmt.body, self.context)]

        if self.stmt.orelse:
            with self.context.block_scope():
                body.extend([parse_body(self.stmt.orelse, self.context)])

        return IRnode.from_list(body)

    def parse_Log(self):
        event = self.stmt._metadata["type"]

        args = [Expr(arg, self.context).ir_node for arg in self.stmt.value.args]

        topic_ir = []
        data_ir = []
        for arg, is_indexed in zip(args, event.indexed):
            if is_indexed:
                topic_ir.append(arg)
            else:
                data_ir.append(arg)

        return events.ir_node_for_log(self.stmt, event, topic_ir, data_ir, self.context)

    def _assert_reason(self, test_expr, msg):
        # from parse_Raise: None passed as the assert condition
        is_raise = test_expr is None

        if isinstance(msg, vy_ast.Name) and msg.id == "UNREACHABLE":
            if is_raise:
                return IRnode.from_list(["invalid"], error_msg="raise unreachable")
            else:
                return IRnode.from_list(
                    ["assert_unreachable", test_expr], error_msg="assert unreachable"
                )

        # set constant so that revert reason str is well behaved
        try:
            tmp = self.context.constancy
            self.context.constancy = Constancy.Constant
            msg_ir = Expr(msg, self.context).ir_node
        finally:
            self.context.constancy = tmp

        msg_ir = wrap_value_for_external_return(msg_ir)
        bufsz = 64 + msg_ir.typ.memory_bytes_required
        buf = self.context.new_internal_variable(get_type_for_exact_size(bufsz))

        # offset of bytes in (bytes,)
        method_id = util.method_id_int("Error(string)")

        # abi encode method_id + bytestring to `buf+32`, then
        # write method_id to `buf` and get out of here
        payload_buf = add_ofst(buf, 32)
        bufsz -= 32  # reduce buffer by size of `method_id` slot
        encoded_length = abi_encode(payload_buf, msg_ir, self.context, bufsz, returns_len=True)
        with encoded_length.cache_when_complex("encoded_len") as (b1, encoded_length):
            revert_seq = [
                "seq",
                ["mstore", buf, method_id],
                ["revert", add_ofst(buf, 28), ["add", 4, encoded_length]],
            ]
            revert_seq = b1.resolve(revert_seq)

        if is_raise:
            ir_node = revert_seq
        else:
            ir_node = ["if", ["iszero", test_expr], revert_seq]
        return IRnode.from_list(ir_node, error_msg="user revert with reason")

    def parse_Assert(self):
        test_expr = Expr.parse_value_expr(self.stmt.test, self.context)

        if self.stmt.msg:
            return self._assert_reason(test_expr, self.stmt.msg)
        else:
            return IRnode.from_list(["assert", test_expr], error_msg="user assert")

    def parse_Raise(self):
        if self.stmt.exc:
            return self._assert_reason(None, self.stmt.exc)
        else:
            return IRnode.from_list(["revert", 0, 0], error_msg="user raise")

    def parse_For(self):
        with self.context.block_scope():
            if self.stmt.get("iter.func.id") == "range":
                return self._parse_For_range()
            else:
                return self._parse_For_list()

    def _parse_For_range(self):
        assert "type" in self.stmt.target.target._metadata
        target_type = self.stmt.target.target._metadata["type"]

        range_call: vy_ast.Call = self.stmt.iter
        assert isinstance(range_call, vy_ast.Call)

        with self.context.range_scope():
            args = [Expr.parse_value_expr(arg, self.context) for arg in range_call.args]
            if len(args) == 1:
                start = IRnode.from_list(0, typ=target_type)
                end = args[0]
            elif len(args) == 2:
                start, end = args
            else:  # pragma: nocover
                raise TypeCheckFailure("unreachable")

            kwargs = {
                s.arg: Expr.parse_value_expr(s.value, self.context) for s in range_call.keywords
            }

        if "bound" in kwargs:
            with end.cache_when_complex("end") as (b1, end):
                # note: the check for rounds<=rounds_bound happens in asm
                # generation for `repeat`.
                clamped_start = clamp_le(start, end, target_type.is_signed)
                rounds = b1.resolve(IRnode.from_list(["sub", end, clamped_start]))
            rounds_bound = kwargs.pop("bound").int_value()
        else:
            rounds = end.int_value() - start.int_value()
            rounds_bound = rounds

        assert len(kwargs) == 0  # sanity check stray keywords

        if rounds_bound < 1:  # pragma: nocover
            raise TypeCheckFailure("unreachable: unchecked 0 bound")

        varname = self.stmt.target.target.id
        i = IRnode.from_list(self.context.fresh_varname("range_ix"), typ=target_type)
        iptr = self.context.new_variable(varname, target_type)

        self.context.forvars[varname] = True

        loop_body = ["seq"]
        # store the current value of i so it is accessible to userland
        loop_body.append(["mstore", iptr, i])
        loop_body.append(parse_body(self.stmt.body, self.context))

        # NOTE: codegen for `repeat` inserts an assertion that
        # (gt rounds_bound rounds). note this also covers the case where
        # rounds < 0.
        # if we ever want to remove that, we need to manually add the assertion
        # where it makes sense.
        ir_node = IRnode.from_list(
            ["repeat", i, start, rounds, rounds_bound, loop_body], error_msg="range() bounds check"
        )
        del self.context.forvars[varname]

        return ir_node

    def _parse_For_list(self):
        with self.context.range_scope():
            iter_list = Expr(self.stmt.iter, self.context).ir_node

        target_type = self.stmt.target.target._metadata["type"]
        assert target_type == iter_list.typ.value_type

        # user-supplied name for loop variable
        varname = self.stmt.target.target.id
        loop_var = IRnode.from_list(
            self.context.new_variable(varname, target_type), typ=target_type, location=MEMORY
        )

        i = IRnode.from_list(self.context.fresh_varname("for_list_ix"), typ=UINT256_T)

        self.context.forvars[varname] = True

        ret = ["seq"]

        # list literal, force it to memory first
        if isinstance(self.stmt.iter, vy_ast.List):
            tmp_list = IRnode.from_list(
                self.context.new_internal_variable(iter_list.typ),
                typ=iter_list.typ,
                location=MEMORY,
            )
            ret.append(make_setter(tmp_list, iter_list))
            iter_list = tmp_list

        # set up the loop variable
        e = get_element_ptr(iter_list, i, array_bounds_check=False)
        body = ["seq", make_setter(loop_var, e), parse_body(self.stmt.body, self.context)]

        repeat_bound = iter_list.typ.count
        if isinstance(iter_list.typ, DArrayT):
            array_len = get_dyn_array_count(iter_list)
        else:
            array_len = repeat_bound

        ret.append(["repeat", i, 0, array_len, repeat_bound, body])

        del self.context.forvars[varname]
        return IRnode.from_list(ret)

    def parse_AugAssign(self):
        target = self._get_target(self.stmt.target)
        right = Expr.parse_value_expr(self.stmt.value, self.context)

        if not target.typ._is_prim_word:
            # because of this check, we do not need to check for
            # make_setter references lhs<->rhs as in parse_Assign -
            # single word load/stores are atomic.
            raise TypeCheckFailure("unreachable")

        with target.cache_when_complex("_loc") as (b, target):
            left = IRnode.from_list(LOAD(target), typ=target.typ)
            new_val = Expr.handle_binop(self.stmt.op, left, right, self.context)
            return b.resolve(STORE(target, new_val))

    def parse_Continue(self):
        return IRnode.from_list("continue")

    def parse_Break(self):
        return IRnode.from_list("break")

    def parse_Return(self):
        ir_val = None
        if self.stmt.value is not None:
            ir_val = Expr(self.stmt.value, self.context).ir_node
        return make_return_stmt(ir_val, self.stmt, self.context)

    def _get_target(self, target):
        _dbg_expr = target

        if isinstance(target, vy_ast.Name) and target.id in self.context.forvars:
            raise TypeCheckFailure(f"Failed constancy check\n{_dbg_expr}")

        if isinstance(target, vy_ast.Tuple):
            target = Expr(target, self.context).ir_node
            for node in target.args:
                if (node.location == STORAGE and self.context.is_constant()) or not node.mutable:
                    raise TypeCheckFailure(f"Failed constancy check\n{_dbg_expr}")
            return target

        target = Expr.parse_pointer_expr(target, self.context)
        if (target.location == STORAGE and self.context.is_constant()) or not target.mutable:
            raise TypeCheckFailure(f"Failed constancy check\n{_dbg_expr}")
        return target


# Parse a statement (usually one line of code but not always)
def parse_stmt(stmt, context):
    return Stmt(stmt, context).ir_node


# check if a function body is "terminated"
# a function is terminated if it ends with a return stmt, OR,
# it ends with an if/else and both branches are terminated.
# (if not, we need to insert a terminator so that the IR is well-formed)
def _is_terminated(code):
    last_stmt = code[-1]

    if last_stmt.is_terminus:
        return True

    if isinstance(last_stmt, vy_ast.If):
        if last_stmt.orelse:
            return _is_terminated(last_stmt.body) and _is_terminated(last_stmt.orelse)
    return False


# codegen a list of statements
def parse_body(code, context, ensure_terminated=False):
    ir_node = ["seq"]
    for stmt in code:
        ir = parse_stmt(stmt, context)
        ir_node.append(ir)

    # force using the return routine / exit_to cleanup for end of function
    if ensure_terminated and context.return_type is None and not _is_terminated(code):
        ir_node.append(parse_stmt(vy_ast.Return(value=None), context))

    # force zerovalent, even last statement
    ir_node.append("pass")  # CMC 2022-01-16 is this necessary?
    return IRnode.from_list(ir_node)



from typing import Dict

from vyper.semantics.types.function import ContractFunctionT


def build_gas_estimates(func_ts: Dict[str, ContractFunctionT]) -> dict:
    # note: `.gas_estimate` is added to ContractFunctionT._ir_info
    # in vyper/semantics/types/function.py
    ret = {}
    for k, v in func_ts.items():
        ret[k] = v._ir_info.gas_estimate
    return ret


def expand_source_map(compressed_map: str) -> list:
    """
    Expand a compressed source map string.

    Arguments
    ---------
    compressed_map : str
        `sourceMap` as generated by the compiler, i.e. "1:1:0;;;;3::2;;;4;"

    Returns
    -------
    List
        Expanded source map as `[[start, length, jump, source id], .. ]`
    """
    source_map: list = [_expand_row(i) if i else None for i in compressed_map.split(";")[:-1]]

    for i, value in enumerate(source_map[1:], 1):
        if value is None:
            source_map[i] = source_map[i - 1][:3] + [None]
            continue
        for x in range(3):
            if source_map[i][x] is None:
                source_map[i][x] = source_map[i - 1][x]

    return source_map


def _expand_row(row):
    result = [None] * 4
    for i, value in enumerate(row.split(":")):
        if value:
            result[i] = value if i == 3 else int(value)
    return result



from vyper.venom.basicblock import IROperand, IRVariable


class StackModel:
    NOT_IN_STACK = object()
    _stack: list[IROperand]

    def __init__(self):
        self._stack = []

    def copy(self):
        new = StackModel()
        new._stack = self._stack.copy()
        return new

    @property
    def height(self) -> int:
        """
        Returns the height of the stack map.
        """
        return len(self._stack)

    def push(self, op: IROperand) -> None:
        """
        Pushes an operand onto the stack map.
        """
        assert isinstance(op, IROperand), f"{type(op)}: {op}"
        self._stack.append(op)

    def pop(self, num: int = 1) -> None:
        del self._stack[len(self._stack) - num :]

    def get_depth(self, op: IROperand, n: int = 1) -> int:
        """
        Returns the depth of the n-th matching operand in the stack map.
        If the operand is not in the stack map, returns NOT_IN_STACK.
        """
        assert isinstance(op, IROperand), f"{type(op)}: {op}"

        for i, stack_op in enumerate(reversed(self._stack)):
            if stack_op.value == op.value:
                if n <= 1:
                    return -i
                else:
                    n -= 1

        return StackModel.NOT_IN_STACK  # type: ignore

    def get_phi_depth(self, phis: list[IRVariable]) -> int:
        """
        Returns the depth of the first matching phi variable in the stack map.
        If the none of the phi operands are in the stack, returns NOT_IN_STACK.
        Asserts that exactly one of phis is found.
        """
        assert isinstance(phis, list)

        ret = StackModel.NOT_IN_STACK
        for i, stack_item in enumerate(reversed(self._stack)):
            if stack_item in phis:
                assert (
                    ret is StackModel.NOT_IN_STACK
                ), f"phi argument is not unique! {phis}, {self._stack}"
                ret = -i

        return ret  # type: ignore

    def peek(self, depth: int) -> IROperand:
        """
        Returns the top of the stack map.
        """
        assert depth is not StackModel.NOT_IN_STACK, "Cannot peek non-in-stack depth"
        return self._stack[depth - 1]

    def poke(self, depth: int, op: IROperand) -> None:
        """
        Pokes an operand at the given depth in the stack map.
        """
        assert depth is not StackModel.NOT_IN_STACK, "Cannot poke non-in-stack depth"
        assert depth <= 0, "Bad depth"
        assert isinstance(op, IROperand), f"{type(op)}: {op}"
        self._stack[depth - 1] = op

    def dup(self, depth: int) -> None:
        """
        Duplicates the operand at the given depth in the stack map.
        """
        assert depth is not StackModel.NOT_IN_STACK, "Cannot dup non-existent operand"
        assert depth <= 0, "Cannot dup positive depth"
        self._stack.append(self.peek(depth))

    def swap(self, depth: int) -> None:
        """
        Swaps the operand at the given depth in the stack map with the top of the stack.
        """
        assert depth is not StackModel.NOT_IN_STACK, "Cannot swap non-existent operand"
        assert depth < 0, "Cannot swap positive depth"
        top = self._stack[-1]
        self._stack[-1] = self._stack[depth - 1]
        self._stack[depth - 1] = top

    def __repr__(self) -> str:
        return f"<StackModel: {self._stack}>"



from vyper.exceptions import TypeMismatch


def test_basic_in_list(get_contract):
    code = """
@external
def testin(x: int128) -> bool:
    y: int128 = 1
    s: int128[4]  = [1, 2, 3, 4]
    if (x + 1) in s:
        return True
    return False
    """

    c = get_contract(code)

    assert c.testin(0) is True
    assert c.testin(1) is True
    assert c.testin(2) is True
    assert c.testin(3) is True
    assert c.testin(4) is False
    assert c.testin(5) is False
    assert c.testin(-1) is False


def test_in_storage_list(get_contract):
    code = """
allowed: int128[10]

@external
def in_test(x: int128) -> bool:
    self.allowed = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    if x in self.allowed:
        return True
    return False
    """

    c = get_contract(code)

    assert c.in_test(1) is True
    assert c.in_test(9) is True
    assert c.in_test(11) is False
    assert c.in_test(-1) is False
    assert c.in_test(32000) is False


def test_in_calldata_list(get_contract):
    code = """
@external
def in_test(x: int128, y: int128[10]) -> bool:
    if x in y:
        return True
    return False
    """

    c = get_contract(code)

    assert c.in_test(1, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) is True
    assert c.in_test(9, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) is True
    assert c.in_test(11, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) is False
    assert c.in_test(-1, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) is False
    assert c.in_test(32000, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]) is False


def test_cmp_in_list(get_contract):
    code = """
@external
def in_test(x: int128) -> bool:
    if x in [9, 7, 6, 5]:
        return True
    return False
    """

    c = get_contract(code)

    assert c.in_test(1) is False
    assert c.in_test(-7) is False
    assert c.in_test(-9) is False
    assert c.in_test(5) is True
    assert c.in_test(7) is True


def test_cmp_not_in_list(get_contract):
    code = """
@external
def in_test(x: int128) -> bool:
    if x not in [9, 7, 6, 5]:
        return True
    return False
    """

    c = get_contract(code)

    assert c.in_test(1) is True
    assert c.in_test(-7) is True
    assert c.in_test(-9) is True
    assert c.in_test(5) is False
    assert c.in_test(7) is False


def test_mixed_in_list(assert_compile_failed, get_contract):
    code = """
@external
def testin() -> bool:
    s: int128[4] = [1, 2, 3, 4]
    if "test" in s:
        return True
    return False
    """
    assert_compile_failed(lambda: get_contract(code), TypeMismatch)


def test_ownership(env, tx_failed, get_contract):
    code = """

owners: address[2]

@deploy
def __init__():
    self.owners[0] = msg.sender

@external
def set_owner(i: int128, new_owner: address):
    assert msg.sender in self.owners
    self.owners[i] = new_owner

@external
def is_owner() -> bool:
    return msg.sender in self.owners
    """
    a1 = env.accounts[1]
    c = get_contract(code)

    assert c.is_owner() is True  # contract creator is owner.
    assert c.is_owner(sender=a1) is False  # no one else is.

    # only an owner may set another owner.
    with tx_failed():
        c.set_owner(1, a1, sender=a1)

    c.set_owner(1, a1)
    assert c.is_owner(sender=a1) is True

    # Owner in place 0 can be replaced.
    c.set_owner(0, a1)
    assert c.is_owner() is False


def test_in_fails_when_types_dont_match(get_contract, tx_failed):
    code = """
@external
def testin(x: address) -> bool:
    s: int128[4] = [1, 2, 3, 4]
    if x in s:
        return True
    return False
"""
    with tx_failed(TypeMismatch):
        get_contract(code)



def test_simple_import(get_contract, make_input_bundle):
    lib1 = """
counter: uint256

@internal
def increment_counter():
    self.counter += 1
    """

    contract = """
import lib

initializes: lib

@external
def increment_counter():
    lib.increment_counter()

@external
def get_counter() -> uint256:
    return lib.counter
    """

    input_bundle = make_input_bundle({"lib.vy": lib1})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.get_counter() == 0
    c.increment_counter()
    assert c.get_counter() == 1


def test_import_namespace(get_contract, make_input_bundle):
    # test what happens when things in current and imported modules share names
    lib = """
counter: uint256

@internal
def increment_counter():
    self.counter += 1
    """

    contract = """
import library as lib

counter: uint256

initializes: lib

@external
def increment_counter():
    self.counter += 1

@external
def increment_lib_counter():
    lib.increment_counter()

@external
def increment_lib_counter2():
    # modify lib.counter directly
    lib.counter += 5

@external
def get_counter() -> uint256:
    return self.counter

@external
def get_lib_counter() -> uint256:
    return lib.counter
    """

    input_bundle = make_input_bundle({"library.vy": lib})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.get_counter() == c.get_lib_counter() == 0

    c.increment_counter()
    assert c.get_counter() == 1
    assert c.get_lib_counter() == 0

    c.increment_lib_counter()
    assert c.get_lib_counter() == 1
    assert c.get_counter() == 1

    c.increment_lib_counter2()
    assert c.get_lib_counter() == 6
    assert c.get_counter() == 1


def test_init_function_side_effects(get_contract, make_input_bundle):
    lib = """
counter: uint256

MY_IMMUTABLE: immutable(uint256)

@deploy
def __init__(initial_value: uint256):
    self.counter = initial_value
    MY_IMMUTABLE = initial_value * 2

@internal
def increment_counter():
    self.counter += 1
    """

    contract = """
import library as lib

counter: public(uint256)

MY_IMMUTABLE: public(immutable(uint256))

initializes: lib

@deploy
def __init__():
    self.counter = 1
    MY_IMMUTABLE = 3
    lib.__init__(5)

@external
def get_lib_counter() -> uint256:
    return lib.counter

@external
def get_lib_immutable() -> uint256:
    return lib.MY_IMMUTABLE
    """

    input_bundle = make_input_bundle({"library.vy": lib})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.counter() == 1
    assert c.MY_IMMUTABLE() == 3
    assert c.get_lib_counter() == 5
    assert c.get_lib_immutable() == 10


def test_indirect_variable_uses(get_contract, make_input_bundle):
    lib1 = """
counter: uint256

MY_IMMUTABLE: immutable(uint256)

@deploy
def __init__(initial_value: uint256):
    self.counter = initial_value
    MY_IMMUTABLE = initial_value * 2

@internal
def increment_counter():
    self.counter += 1
    """
    lib2 = """
import lib1

uses: lib1

@internal
def get_lib1_counter() -> uint256:
    return lib1.counter

@internal
def get_lib1_my_immutable() -> uint256:
    return lib1.MY_IMMUTABLE
    """

    contract = """
import lib1
import lib2

initializes: lib1
initializes: lib2[lib1 := lib1]

@deploy
def __init__():
    lib1.__init__(5)

@external
def get_storage_via_lib1() -> uint256:
    return lib1.counter

@external
def get_immutable_via_lib1() -> uint256:
    return lib1.MY_IMMUTABLE

@external
def get_storage_via_lib2() -> uint256:
    return lib2.get_lib1_counter()

@external
def get_immutable_via_lib2() -> uint256:
    return lib2.get_lib1_my_immutable()
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.get_storage_via_lib1() == c.get_storage_via_lib2() == 5
    assert c.get_immutable_via_lib1() == c.get_immutable_via_lib2() == 10


def test_uses_already_initialized(get_contract, make_input_bundle):
    lib1 = """
counter: uint256
MY_IMMUTABLE: immutable(uint256)

@deploy
def __init__(initial_value: uint256):
    self.counter = initial_value * 2
    MY_IMMUTABLE = initial_value * 3

@internal
def increment_counter():
    self.counter += 1
    """
    lib2 = """
import lib1

initializes: lib1

@deploy
def __init__():
    lib1.__init__(5)

@internal
def get_lib1_counter() -> uint256:
    return lib1.counter

@internal
def get_lib1_my_immutable() -> uint256:
    return lib1.MY_IMMUTABLE
    """

    contract = """
import lib1
import lib2

uses: lib1
initializes: lib2

@deploy
def __init__():
    lib2.__init__()

@external
def get_storage_via_lib1() -> uint256:
    return lib1.counter

@external
def get_immutable_via_lib1() -> uint256:
    return lib1.MY_IMMUTABLE

@external
def get_storage_via_lib2() -> uint256:
    return lib2.get_lib1_counter()

@external
def get_immutable_via_lib2() -> uint256:
    return lib2.get_lib1_my_immutable()
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.get_storage_via_lib1() == c.get_storage_via_lib2() == 10
    assert c.get_immutable_via_lib1() == c.get_immutable_via_lib2() == 15


def test_import_complex_types(get_contract, make_input_bundle):
    lib1 = """
an_array: uint256[3]
a_hashmap: HashMap[address, HashMap[uint256, uint256]]

@internal
def set_array_value(ix: uint256, new_value: uint256):
    self.an_array[ix] = new_value

@internal
def set_hashmap_value(ix0: address, ix1: uint256, new_value: uint256):
    self.a_hashmap[ix0][ix1] = new_value
    """

    contract = """
import lib

initializes: lib

@external
def do_things():
    lib.set_array_value(1, 5)
    lib.set_hashmap_value(msg.sender, 6, 100)

@external
def get_array_value(ix: uint256) -> uint256:
    return lib.an_array[ix]

@external
def get_hashmap_value(ix: uint256) -> uint256:
    return lib.a_hashmap[msg.sender][ix]
    """

    input_bundle = make_input_bundle({"lib.vy": lib1})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.get_array_value(0) == 0
    assert c.get_hashmap_value(0) == 0
    c.do_things()

    assert c.get_array_value(0) == 0
    assert c.get_hashmap_value(0) == 0
    assert c.get_array_value(1) == 5
    assert c.get_hashmap_value(6) == 100



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import (
    ImmutableViolation,
    InterfaceViolation,
    NamespaceCollision,
    StructureException,
)

from .helpers import NONREENTRANT_NOTE


def test_exports_no_uses(make_input_bundle):
    lib1 = """
counter: uint256

@external
def get_counter() -> uint256:
    self.counter += 1
    return self.counter
    """
    main = """
import lib1
exports: lib1.get_counter
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value.hint == expected_hint


def test_exports_no_uses_variable(make_input_bundle):
    lib1 = """
counter: public(uint256)
    """
    main = """
import lib1
exports: lib1.counter
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(ImmutableViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value._message == "Cannot access `lib1` state!" + NONREENTRANT_NOTE

    expected_hint = "add `uses: lib1` or `initializes: lib1` as a "
    expected_hint += "top-level statement to your contract"
    assert e.value.hint == expected_hint


def test_exports_uses_variable(make_input_bundle):
    lib1 = """
counter: public(uint256)
    """
    main = """
import lib1

exports: lib1.counter
initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    assert compile_code(main, input_bundle=input_bundle) is not None


def test_exports_uses(make_input_bundle):
    lib1 = """
counter: uint256

@external
def get_counter() -> uint256:
    self.counter += 1
    return self.counter
    """
    main = """
import lib1

exports: lib1.get_counter
initializes: lib1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    assert compile_code(main, input_bundle=input_bundle) is not None


# test that exporting can satisfy an implements constraint
# use a mix of public variables and functions
def test_exports_implements(make_input_bundle):
    token_interface = """
@external
@view
def totalSupply() -> uint256:
    ...

@external
@view
def balanceOf(addr: address) -> uint256:
    ...

@external
def transfer(receiver: address, amount: uint256):
    ...
    """
    lib1 = """
import itoken

implements: itoken

@deploy
def __init__(initial_supply: uint256):
    self.totalSupply = initial_supply
    self.balanceOf[msg.sender] = initial_supply

totalSupply: public(uint256)
balanceOf: public(HashMap[address, uint256])

@external
def transfer(receiver: address, amount: uint256):
    self.balanceOf[msg.sender] -= amount
    self.balanceOf[receiver] += amount
    """
    main = """
import tokenlib
import itoken

implements: itoken
exports: (tokenlib.totalSupply, tokenlib.balanceOf, tokenlib.transfer)

initializes: tokenlib

@deploy
def __init__():
    tokenlib.__init__(100_000_000)
    """
    input_bundle = make_input_bundle({"tokenlib.vy": lib1, "itoken.vyi": token_interface})
    assert compile_code(main, input_bundle=input_bundle) is not None


# test that exporting can satisfy an implements constraint
# use a mix of local and imported functions
def test_exports_implements2(make_input_bundle):
    ifoobar = """
@external
def foo():
    ...

@external
def bar():
    ...
    """
    lib1 = """
import ifoobar

implements: ifoobar

counter: uint256

@external
def foo():
    pass

@external
def bar():
    self.counter += 1
    """
    main = """
import lib1
import ifoobar

implements: ifoobar
exports: lib1.foo

initializes: lib1

# for fun, export a different function with the same name
@external
def bar():
    lib1.counter += 2
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "ifoobar.vyi": ifoobar})
    assert compile_code(main, input_bundle=input_bundle) is not None


def test_function_name_collisions(make_input_bundle):
    lib1 = """
@external
def foo():
    pass
    """
    main = """
import lib1

exports: lib1.foo

@external
def foo():
    x: uint256 = 12345
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(NamespaceCollision) as e:
        # TODO: make the error message reference the export
        compile_code(main, contract_path="main.vy", input_bundle=input_bundle)

    assert e.value._message == "Member 'foo' already exists in self"

    assert e.value.annotations[0].lineno == 4
    assert e.value.annotations[0].node_source_code == "lib1.foo"
    assert e.value.annotations[0].module_node.path == "main.vy"

    assert e.value.prev_decl.lineno == 7
    assert e.value.prev_decl.node_source_code.startswith("def foo():")
    assert e.value.prev_decl.module_node.path == "main.vy"


def test_duplicate_exports(make_input_bundle):
    lib1 = """
@external
def foo():
    pass

@external
def bar():
    pass
    """
    main = """
import lib1

exports: lib1.foo
exports: lib1.bar
exports: lib1.foo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(StructureException) as e:
        # TODO: make the error message reference the export
        compile_code(main, contract_path="main.vy", input_bundle=input_bundle)

    assert e.value._message == "already exported!"

    assert e.value.annotations[0].lineno == 6
    assert e.value.annotations[0].node_source_code == "lib1.foo"
    assert e.value.annotations[0].module_node.path == "main.vy"

    assert e.value.prev_decl.lineno == 4
    assert e.value.prev_decl.node_source_code == "lib1.foo"
    assert e.value.prev_decl.module_node.path == "main.vy"


def test_duplicate_exports_tuple(make_input_bundle):
    lib1 = """
@external
def foo():
    pass

@external
def bar():
    pass
    """
    main = """
import lib1

exports: (lib1.foo, lib1.bar, lib1.foo)
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(StructureException) as e:
        # TODO: make the error message reference the export
        compile_code(main, contract_path="main.vy", input_bundle=input_bundle)

    assert e.value._message == "already exported!"

    assert e.value.annotations[0].lineno == 4
    assert e.value.annotations[0].col_offset == 30
    assert e.value.annotations[0].node_source_code == "lib1.foo"
    assert e.value.annotations[0].module_node.path == "main.vy"

    assert e.value.prev_decl.lineno == 4
    assert e.value.prev_decl.col_offset == 10
    assert e.value.prev_decl.node_source_code == "lib1.foo"
    assert e.value.prev_decl.module_node.path == "main.vy"


def test_duplicate_exports_tuple2(make_input_bundle):
    lib1 = """
@external
def foo():
    pass

@external
def bar():
    pass
    """
    main = """
import lib1

exports: lib1.foo
exports: (lib1.bar, lib1.foo)
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(StructureException) as e:
        # TODO: make the error message reference the export
        compile_code(main, contract_path="main.vy", input_bundle=input_bundle)

    assert e.value._message == "already exported!"

    assert e.value.annotations[0].lineno == 5
    assert e.value.annotations[0].col_offset == 20
    assert e.value.annotations[0].node_source_code == "lib1.foo"
    assert e.value.annotations[0].module_node.path == "main.vy"

    assert e.value.prev_decl.lineno == 4
    assert e.value.prev_decl.col_offset == 9
    assert e.value.prev_decl.node_source_code == "lib1.foo"
    assert e.value.prev_decl.module_node.path == "main.vy"


def test_interface_export_collision(make_input_bundle):
    main = """
import lib1

exports: lib1.__interface__
exports: lib1.bar
    """
    lib1 = """
@external
def bar() -> uint256:
    return 1
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})
    with pytest.raises(StructureException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "already exported!"


def test_no_export_missing_function(make_input_bundle):
    ifoo = """
@external
def do_xyz():
    ...
    """
    lib1 = """
import ifoo

@external
@view
def bar() -> uint256:
    return 1
    """
    main = """
import lib1

exports: lib1.ifoo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "ifoo.vyi": ifoo})
    with pytest.raises(InterfaceViolation) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "requested `lib1.ifoo` but `lib1` does not implement `lib1.ifoo`!"


def test_no_export_unimplemented_interface(make_input_bundle):
    ifoo = """
@external
def do_xyz():
    ...
    """
    lib1 = """
import ifoo

# technically implements ifoo, but missing `implements: ifoo`

@external
def do_xyz():
    pass
    """
    main = """
import lib1

exports: lib1.ifoo
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "ifoo.vyi": ifoo})
    with pytest.raises(InterfaceViolation) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "requested `lib1.ifoo` but `lib1` does not implement `lib1.ifoo`!"


def test_export_selector_conflict(make_input_bundle):
    ifoo = """
@external
def gsf():
    ...
    """
    lib1 = """
import ifoo

@external
def gsf():
    pass

@external
@view
def tgeo() -> uint256:
    return 1
    """
    main = """
import lib1

exports: (lib1.ifoo, lib1.tgeo)
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "ifoo.vyi": ifoo})
    with pytest.raises(StructureException) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "Methods produce colliding method ID `0x67e43e43`: gsf(), tgeo()"


def test_export_different_return_type(make_input_bundle):
    ifoo = """
@external
def foo() -> uint256:
    ...
    """
    lib1 = """
import ifoo

foo: public(int256)

@deploy
def __init__():
    self.foo = -1
    """
    main = """
import lib1

initializes: lib1

exports: lib1.ifoo

@deploy
def __init__():
    lib1.__init__()
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "ifoo.vyi": ifoo})
    with pytest.raises(InterfaceViolation) as e:
        compile_code(main, input_bundle=input_bundle)
    assert e.value._message == "requested `lib1.ifoo` but `lib1` does not implement `lib1.ifoo`!"



from typing import Any, Callable


def levenshtein_norm(source: str, target: str) -> float:
    """Calculates the normalized Levenshtein distance between two string
    arguments. The result will be a float in the range [0.0, 1.0], with 1.0
    signifying the biggest possible distance between strings with these lengths

    From jazzband/docopt-ng
    https://github.com/jazzband/docopt-ng/blob/bbed40a2335686d2e14ac0e6c3188374dc4784da/docopt.py
    """

    # Compute Levenshtein distance using helper function. The max is always
    # just the length of the longer string, so this is used to normalize result
    # before returning it
    distance = levenshtein(source, target)
    return float(distance) / max(len(source), len(target))


def levenshtein(source: str, target: str) -> int:
    """Computes the Levenshtein
    (https://en.wikipedia.org/wiki/Levenshtein_distance)
    and restricted Damerau-Levenshtein
    (https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance)
    distances between two Unicode strings with given lengths using the
    Wagner-Fischer algorithm
    (https://en.wikipedia.org/wiki/Wagner%E2%80%93Fischer_algorithm).
    These distances are defined recursively, since the distance between two
    strings is just the cost of adjusting the last one or two characters plus
    the distance between the prefixes that exclude these characters (e.g. the
    distance between "tester" and "tested" is 1 + the distance between "teste"
    and "teste"). The Wagner-Fischer algorithm retains this idea but eliminates
    redundant computations by storing the distances between various prefixes in
    a matrix that is filled in iteratively.

    From jazzband/docopt-ng
    https://github.com/jazzband/docopt-ng/blob/bbed40a2335686d2e14ac0e6c3188374dc4784da/docopt.py
    """

    # Create matrix of correct size (this is s_len + 1 * t_len + 1 so that the
    # empty prefixes "" can also be included). The leftmost column represents
    # transforming various source prefixes into an empty string, which can
    # always be done by deleting all characters in the respective prefix, and
    # the top row represents transforming the empty string into various target
    # prefixes, which can always be done by inserting every character in the
    # respective prefix. The ternary used to build the list should ensure that
    # this row and column are now filled correctly
    s_range = range(len(source) + 1)
    t_range = range(len(target) + 1)
    matrix = [[(i if j == 0 else j) for j in t_range] for i in s_range]

    # Iterate through rest of matrix, filling it in with Levenshtein
    # distances for the remaining prefix combinations
    for i in s_range[1:]:
        for j in t_range[1:]:
            # Applies the recursive logic outlined above using the values
            # stored in the matrix so far. The options for the last pair of
            # characters are deletion, insertion, and substitution, which
            # amount to dropping the source character, the target character,
            # or both and then calculating the distance for the resulting
            # prefix combo. If the characters at this point are the same, the
            # situation can be thought of as a free substitution
            del_dist = matrix[i - 1][j] + 1
            ins_dist = matrix[i][j - 1] + 1
            sub_trans_cost = 0 if source[i - 1] == target[j - 1] else 1
            sub_dist = matrix[i - 1][j - 1] + sub_trans_cost

            # Choose option that produces smallest distance
            matrix[i][j] = min(del_dist, ins_dist, sub_dist)

    # At this point, the matrix is full, and the biggest prefixes are just the
    # strings themselves, so this is the desired distance
    return matrix[len(source)][len(target)]


def get_levenshtein_error_suggestions(*args, **kwargs) -> Callable:
    return lambda: _get_levenshtein_error_suggestions(*args, **kwargs)


def _get_levenshtein_error_suggestions(
    key: str, namespace: dict[str, Any], threshold: float
) -> str:
    """
    Generate an error message snippet for the suggested closest values in the provided namespace
    with the shortest normalized Levenshtein distance from the given key if that distance
    is below the threshold. Otherwise, return an empty string.

    As a heuristic, the threshold value is inversely correlated to the size of the namespace.
    For a small namespace (e.g. struct members), the threshold value can be the maximum of
    1.0 since the key must be one of the defined struct members. For a large namespace
    (e.g. types, builtin functions and state variables), the threshold value should be lower
    to ensure the matches are relevant.

    :param key: A string of the identifier being accessed
    :param namespace: A dictionary of the possible identifiers
    :param threshold: A floating value between 0.0 and 1.0

    :return: The error message snippet if the Levenshtein value is below the threshold,
        or an empty string.
    """

    if key is None or key == "":
        return ""

    distances = sorted([(i, levenshtein_norm(key, i)) for i in namespace], key=lambda k: k[1])
    if len(distances) > 0 and distances[0][1] <= threshold:
        if len(distances) > 1 and distances[1][1] <= threshold:
            return f"Did you mean '{distances[0][0]}', or maybe '{distances[1][0]}'?"
        return f"Did you mean '{distances[0][0]}'?"
    return ""



#!/usr/bin/env python3

import json
from copy import deepcopy

import pytest

from vyper.cli.vyper_json import _parse_args
from vyper.exceptions import JSONError

FOO_CODE = """
import contracts.ibar as IBar

@external
def foo(a: address) -> bool:
    return extcall IBar(a).bar(1)
"""

BAR_CODE = """
@external
def bar(a: uint256) -> bool:
    return True
"""

BAR_ABI = [
    {
        "name": "bar",
        "outputs": [{"type": "bool", "name": "out"}],
        "inputs": [{"type": "uint256", "name": "a"}],
        "stateMutability": "nonpayable",
        "type": "function",
    }
]

INPUT_JSON = {
    "language": "Vyper",
    "sources": {
        "contracts/foo.vy": {"content": FOO_CODE},
        "contracts/bar.vy": {"content": BAR_CODE},
    },
    "interfaces": {"contracts/ibar.json": {"abi": BAR_ABI}},
    "settings": {"outputSelection": {"*": ["*"]}},
}


def _no_errors(output_json):
    return "errors" not in output_json or not any(
        err["severity"] == "error" for err in output_json["errors"]
    )


def test_to_stdout(tmp_path, capfd):
    path = tmp_path.joinpath("input.json")
    with path.open("w") as fp:
        json.dump(INPUT_JSON, fp)
    _parse_args([path.absolute().as_posix()])
    out, _ = capfd.readouterr()
    output_json = json.loads(out)
    assert _no_errors(output_json), (INPUT_JSON, output_json)
    assert "contracts/foo.vy" in output_json["sources"]
    assert "contracts/bar.vy" in output_json["sources"]


def test_to_file(tmp_path):
    path = tmp_path.joinpath("input.json")
    with path.open("w") as fp:
        json.dump(INPUT_JSON, fp)
    output_path = tmp_path.joinpath("output.json")
    _parse_args([path.absolute().as_posix(), "-o", output_path.absolute().as_posix()])
    assert output_path.exists()
    with output_path.open() as fp:
        output_json = json.load(fp)
    assert _no_errors(output_json), (INPUT_JSON, output_json)
    assert "contracts/foo.vy" in output_json["sources"]
    assert "contracts/bar.vy" in output_json["sources"]


def test_pretty_json(tmp_path, capfd):
    path = tmp_path.joinpath("input.json")
    with path.open("w") as fp:
        json.dump(INPUT_JSON, fp)
    _parse_args([path.absolute().as_posix()])
    out1, _ = capfd.readouterr()
    _parse_args([path.absolute().as_posix(), "--pretty-json"])
    out2, _ = capfd.readouterr()
    assert len(out2) > len(out1)
    assert json.loads(out1) == json.loads(out2)


def test_traceback(tmp_path, capfd):
    path = tmp_path.joinpath("input.json")
    input_json = deepcopy(INPUT_JSON)
    del input_json["sources"]
    with path.open("w") as fp:
        json.dump(input_json, fp)
    _parse_args([path.absolute().as_posix()])
    out, _ = capfd.readouterr()
    output_json = json.loads(out)
    assert not _no_errors(output_json)
    with pytest.raises(JSONError):
        _parse_args([path.absolute().as_posix(), "--traceback"])



from vyper.venom.analysis.analysis import IRAnalysesCache
from vyper.venom.analysis.cfg import CFGAnalysis
from vyper.venom.context import IRContext
from vyper.venom.function import IRBasicBlock, IRLabel
from vyper.venom.passes.normalization import NormalizationPass


def test_multi_entry_block_1():
    ctx = IRContext()
    fn = ctx.create_function("__global")

    finish_label = IRLabel("finish")
    target_label = IRLabel("target")
    block_1_label = IRLabel("block_1", fn)

    bb = fn.get_basic_block()
    op = bb.append_instruction("store", 10)
    acc = bb.append_instruction("add", op, op)
    bb.append_instruction("jnz", acc, finish_label, block_1_label)

    block_1 = IRBasicBlock(block_1_label, fn)
    fn.append_basic_block(block_1)
    acc = block_1.append_instruction("add", acc, op)
    op = block_1.append_instruction("store", 10)
    block_1.append_instruction("mstore", acc, op)
    block_1.append_instruction("jnz", acc, finish_label, target_label)

    target_bb = IRBasicBlock(target_label, fn)
    fn.append_basic_block(target_bb)
    target_bb.append_instruction("mul", acc, acc)
    target_bb.append_instruction("jmp", finish_label)

    finish_bb = IRBasicBlock(finish_label, fn)
    fn.append_basic_block(finish_bb)
    finish_bb.append_instruction("stop")

    ac = IRAnalysesCache(fn)
    ac.request_analysis(CFGAnalysis)
    assert not fn.normalized, "CFG should not be normalized"

    NormalizationPass(ac, fn).run_pass()

    assert fn.normalized, "CFG should be normalized"

    finish_bb = fn.get_basic_block(finish_label.value)
    cfg_in = list(finish_bb.cfg_in)
    assert cfg_in[0].label.value == "target", "Should contain target"
    assert cfg_in[1].label.value == "__global_split_finish", "Should contain __global_split_finish"
    assert cfg_in[2].label.value == "block_1_split_finish", "Should contain block_1_split_finish"


# more complicated one
def test_multi_entry_block_2():
    ctx = IRContext()
    fn = ctx.create_function("__global")

    finish_label = IRLabel("finish")
    target_label = IRLabel("target")
    block_1_label = IRLabel("block_1", fn)
    block_2_label = IRLabel("block_2", fn)

    bb = fn.get_basic_block()
    op = bb.append_instruction("store", 10)
    acc = bb.append_instruction("add", op, op)
    bb.append_instruction("jnz", acc, finish_label, block_1_label)

    block_1 = IRBasicBlock(block_1_label, fn)
    fn.append_basic_block(block_1)
    acc = block_1.append_instruction("add", acc, op)
    op = block_1.append_instruction("store", 10)
    block_1.append_instruction("mstore", acc, op)
    block_1.append_instruction("jnz", acc, target_label, finish_label)

    block_2 = IRBasicBlock(block_2_label, fn)
    fn.append_basic_block(block_2)
    acc = block_2.append_instruction("add", acc, op)
    op = block_2.append_instruction("store", 10)
    block_2.append_instruction("mstore", acc, op)
    # switch the order of the labels, for fun and profit
    block_2.append_instruction("jnz", acc, finish_label, target_label)

    target_bb = IRBasicBlock(target_label, fn)
    fn.append_basic_block(target_bb)
    target_bb.append_instruction("mul", acc, acc)
    target_bb.append_instruction("jmp", finish_label)

    finish_bb = IRBasicBlock(finish_label, fn)
    fn.append_basic_block(finish_bb)
    finish_bb.append_instruction("stop")

    ac = IRAnalysesCache(fn)
    ac.request_analysis(CFGAnalysis)
    assert not fn.normalized, "CFG should not be normalized"

    NormalizationPass(ac, fn).run_pass()

    assert fn.normalized, "CFG should be normalized"

    finish_bb = fn.get_basic_block(finish_label.value)
    cfg_in = list(finish_bb.cfg_in)
    assert cfg_in[0].label.value == "target", "Should contain target"
    assert cfg_in[1].label.value == "__global_split_finish", "Should contain __global_split_finish"
    assert cfg_in[2].label.value == "block_1_split_finish", "Should contain block_1_split_finish"


def test_multi_entry_block_with_dynamic_jump():
    ctx = IRContext()
    fn = ctx.create_function("__global")

    finish_label = IRLabel("finish")
    target_label = IRLabel("target")
    block_1_label = IRLabel("block_1", fn)

    bb = fn.get_basic_block()
    op = bb.append_instruction("store", 10)
    acc = bb.append_instruction("add", op, op)
    bb.append_instruction("djmp", acc, finish_label, block_1_label)

    block_1 = IRBasicBlock(block_1_label, fn)
    fn.append_basic_block(block_1)
    acc = block_1.append_instruction("add", acc, op)
    op = block_1.append_instruction("store", 10)
    block_1.append_instruction("mstore", acc, op)
    block_1.append_instruction("jnz", acc, finish_label, target_label)

    target_bb = IRBasicBlock(target_label, fn)
    fn.append_basic_block(target_bb)
    target_bb.append_instruction("mul", acc, acc)
    target_bb.append_instruction("jmp", finish_label)

    finish_bb = IRBasicBlock(finish_label, fn)
    fn.append_basic_block(finish_bb)
    finish_bb.append_instruction("stop")

    ac = IRAnalysesCache(fn)
    ac.request_analysis(CFGAnalysis)
    assert not fn.normalized, "CFG should not be normalized"

    NormalizationPass(ac, fn).run_pass()
    assert fn.normalized, "CFG should be normalized"

    finish_bb = fn.get_basic_block(finish_label.value)
    cfg_in = list(finish_bb.cfg_in)
    assert cfg_in[0].label.value == "target", "Should contain target"
    assert cfg_in[1].label.value == "__global_split_finish", "Should contain __global_split_finish"
    assert cfg_in[2].label.value == "block_1_split_finish", "Should contain block_1_split_finish"



def test_is_contract(env, get_contract):
    contract_1 = """
@external
def foo(arg1: address) -> bool:
    result: bool = arg1.is_contract
    return result
"""

    contract_2 = """
@external
def foo(arg1: address) -> bool:
    return arg1.is_contract
"""
    a0, a1 = env.accounts[:2]
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.foo(c1.address) is True
    assert c1.foo(c2.address) is True
    assert c1.foo(a1) is False
    assert c1.foo(a0) is False
    assert c2.foo(c1.address) is True
    assert c2.foo(c2.address) is True
    assert c2.foo(a1) is False
    assert c2.foo(a0) is False
